from functools import partial

import pytest
import sympy

from pharmpy.model import (
    Assignment,
    DataInfo,
    Model,
    NormalDistribution,
    Parameter,
    Parameters,
    RandomVariables,
    Statements,
)
from pharmpy.modeling import (
    add_effect_compartment,
    add_indirect_effect,
    add_metabolite,
    add_peripheral_compartment,
    calculate_epsilon_gradient_expression,
    calculate_eta_gradient_expression,
    cleanup_model,
    create_basic_pk_model,
    display_odes,
    get_dv_symbol,
    get_individual_parameters,
    get_individual_prediction_expression,
    get_observation_expression,
    get_parameter_rv,
    get_pd_parameters,
    get_pk_parameters,
    get_population_prediction_expression,
    get_rv_parameters,
    greekify_model,
    has_random_effect,
    is_linearized,
    is_real,
    make_declarative,
    mu_reference_model,
    read_model_from_string,
    set_direct_effect,
    simplify_expression,
    solve_ode_system,
)


def s(x):
    return sympy.Symbol(x)


def test_get_observation_expression(testdata, load_model_for_test):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno_real_linbase.mod')
    expr = get_observation_expression(model)
    assert expr == s('D_EPSETA1_2') * s('EPS_1') * (s('ETA_2') - s('OETA2')) + s('D_ETA1') * (
        s('ETA_1') - s('OETA1')
    ) + s('D_ETA2') * (s('ETA_2') - s('OETA2')) + s('EPS_1') * (
        s('D_EPS1') + s('D_EPSETA1_1') * (s('ETA_1') - s('OETA1'))
    ) + s(
        'OPRED'
    )


def test_get_individual_prediction_expression(testdata, load_model_for_test):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno_real_linbase.mod')
    expr = get_individual_prediction_expression(model)
    assert expr == s('D_ETA1') * (s('ETA_1') - s('OETA1')) + s('D_ETA2') * (
        s('ETA_2') - s('OETA2')
    ) + s('OPRED')


def test_get_population_prediction_expression(testdata, load_model_for_test):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno_real_linbase.mod')
    expr = get_population_prediction_expression(model)
    assert expr == -s('D_ETA1') * s('OETA1') - s('D_ETA2') * s('OETA2') + s('OPRED')


def test_calculate_eta_gradient_expression(testdata, load_model_for_test):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno_real_linbase.mod')
    expr = calculate_eta_gradient_expression(model)
    assert expr == [s('D_ETA1'), s('D_ETA2')]


def test_calculate_epsilon_gradient_expression(testdata, load_model_for_test):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno_real_linbase.mod')
    expr = calculate_epsilon_gradient_expression(model)
    assert expr == [
        s('D_EPS1')
        + s('D_EPSETA1_1') * (s('ETA_1') - s('OETA1'))
        + s('D_EPSETA1_2') * (s('ETA_2') - s('OETA2'))
    ]


def test_mu_reference_model_full():
    code = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV FA1 FA2
$SUBROUTINE ADVAN1 TRANS2
$PK
TVCL=THETA(1)*WGT
TVV=THETA(2)*WGT
CL=TVCL*EXP(ETA(1))
V=TVV*EXP(ETA(2))
S1=V
$ERROR
Y=F+F*EPS(1)
$THETA (0,0.00469307) ; PTVCL
$THETA (0,1.00916) ; PTVV
$OMEGA DIAGONAL(2)
0.0309626 ; IVCL
0.031128 ; IVV
$SIGMA 0.013241 ;sigma
$ESTIMATION METHOD=1 INTERACTION
"""
    model = read_model_from_string(code)
    model = mu_reference_model(model)
    correct = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV FA1 FA2
$SUBROUTINE ADVAN1 TRANS2
$PK
TVCL=THETA(1)*WGT
TVV=THETA(2)*WGT
MU_1 = LOG(TVCL)
CL = EXP(MU_1 + ETA(1))
MU_2 = LOG(TVV)
V = EXP(MU_2 + ETA(2))
S1=V
$ERROR
Y=F+F*EPS(1)
$THETA (0,0.00469307) ; PTVCL
$THETA (0,1.00916) ; PTVV
$OMEGA DIAGONAL(2)
0.0309626 ; IVCL
0.031128 ; IVV
$SIGMA 0.013241 ;sigma
$ESTIMATION METHOD=1 INTERACTION
"""
    assert model.model_code == correct


@pytest.mark.usefixtures('testdata')
@pytest.mark.parametrize(
    'statements,correct',
    [
        (
            [Assignment.create('CL', s('THETA(1)') + s('ETA(1)'))],
            [
                Assignment.create('mu_1', s('THETA(1)')),
                Assignment.create('CL', s('mu_1') + s('ETA(1)')),
            ],
        ),
        (
            [
                Assignment.create(
                    'CL', s('THETA(1)') * s('AGE') ** s('THETA(2)') * sympy.exp(s('ETA(1)'))
                ),
                Assignment.create('V', s('THETA(3)') * sympy.exp(s('ETA(2)'))),
            ],
            [
                Assignment.create('mu_1', sympy.log(s('THETA(1)') * s('AGE') ** s('THETA(2)'))),
                Assignment.create('CL', sympy.exp(s('mu_1') + s('ETA(1)'))),
                Assignment.create('mu_2', sympy.log(s('THETA(3)'))),
                Assignment.create('V', sympy.exp(s('mu_2') + s('ETA(2)'))),
            ],
        ),
        (
            [Assignment.create('CL', s('THETA(1)') + s('ETA(1)') + s('ETA(2)'))],
            [Assignment.create('CL', s('THETA(1)') + s('ETA(1)') + s('ETA(2)'))],
        ),
    ],
)
def test_mu_reference_model_generic(statements, correct):
    model = Model()
    datainfo = DataInfo.create(['AGE'])
    eta1 = NormalDistribution.create('ETA(1)', 'iiv', 0, s('omega1'))
    eta2 = NormalDistribution.create('ETA(2)', 'iiv', 0, s('omega2'))
    rvs = RandomVariables.create([eta1, eta2])
    th1 = Parameter('THETA(1)', 2, lower=1)
    th2 = Parameter('THETA(2)', 2, lower=1)
    th3 = Parameter('THETA(3)', 2, lower=1)
    params = Parameters((th1, th2, th3))
    model = model.replace(
        statements=Statements(statements),
        parameters=params,
        random_variables=rvs,
        datainfo=datainfo,
    )
    model = mu_reference_model(model)
    assert model.statements == Statements(correct)


def test_simplify_expression():
    model = Model()
    x = sympy.Symbol('x')
    y = sympy.Symbol('y')

    p1 = Parameter('x', 3)
    p2 = Parameter('y', 9, fix=True)
    pset = Parameters((p1, p2))
    model = model.replace(parameters=pset)
    assert simplify_expression(model, x * y) == 9.0 * x

    p1 = Parameter('x', 3, lower=0.001)
    p2 = Parameter('y', 9)
    pset = Parameters((p1, p2))
    model = model.replace(parameters=pset)
    assert simplify_expression(model, abs(x)) == x

    p1 = Parameter('x', 3, lower=0)
    p2 = Parameter('y', 9)
    pset = Parameters((p1, p2))
    model = model.replace(parameters=pset)
    assert simplify_expression(model, sympy.Piecewise((2, sympy.Ge(x, 0)), (56, True))) == 2

    p1 = Parameter('x', -3, upper=-1)
    p2 = Parameter('y', 9)
    pset = Parameters((p1, p2))
    model = model.replace(parameters=pset)
    assert simplify_expression(model, abs(x)) == -x

    p1 = Parameter('x', -3, upper=0)
    p2 = Parameter('y', 9)
    pset = Parameters((p1, p2))
    model = model.replace(parameters=pset)
    assert simplify_expression(model, sympy.Piecewise((2, sympy.Le(x, 0)), (56, True))) == 2

    p1 = Parameter('x', 3)
    p2 = Parameter('y', 9)
    pset = Parameters((p1, p2))
    model = model.replace(parameters=pset)
    assert simplify_expression(model, x * y) == x * y


def test_solve_ode_system(pheno):
    model = solve_ode_system(pheno)
    assert sympy.Symbol('t') in model.statements[8].free_symbols


def test_make_declarative(pheno):
    model = make_declarative(pheno)
    assert model.statements[3].expression == sympy.Piecewise(
        (s('WGT') * s('PTVV') * (s('THETA_3') + 1), sympy.Lt(s('APGR'), 5)),
        (s('WGT') * s('PTVV'), True),
    )


def test_cleanup_model(pheno):
    model = cleanup_model(pheno)
    assert model.statements.after_odes[1].symbol != s('W')


def test_greekify_model(pheno):
    model = cleanup_model(pheno)
    model = greekify_model(model)
    assert s('theta_1') in model.statements[2].free_symbols


@pytest.mark.parametrize(
    ('model_path', 'kind', 'expected'),
    (
        ('nonmem/pheno.mod', 'all', ['CL', 'V']),
        ('nonmem/pheno_real.mod', 'all', ['CL', 'V']),
        ('nonmem/models/mox1.mod', 'all', ['CL', 'KA', 'V']),
        ('nonmem/models/mox2.mod', 'all', ['CL', 'MAT', 'VC']),
        ('nonmem/models/mox_2comp.mod', 'all', ['K', 'K12', 'K21']),
        ('nonmem/models/pef.mod', 'all', ['K']),
        ('nonmem/models/pheno_advan3_trans1.mod', 'all', ['K', 'K12', 'K21']),
        ('nonmem/models/pheno_noifs.mod', 'all', ['CL', 'V']),
        ('nonmem/models/pheno_trans1.mod', 'all', ['K']),
        ('nonmem/secondary_parameters/run1.mod', 'all', ['CL', 'V']),
        ('nonmem/secondary_parameters/run2.mod', 'all', ['CL', 'V']),
        ('nonmem/qa/iov.mod', 'all', ['CL', 'V']),
        ('nonmem/pheno.mod', 'absorption', []),
        ('nonmem/pheno_real.mod', 'absorption', []),
        ('nonmem/models/mox1.mod', 'absorption', ['KA']),
        ('nonmem/models/mox2.mod', 'absorption', ['MAT']),
        ('nonmem/models/mox_2comp.mod', 'absorption', []),
        ('nonmem/models/pef.mod', 'absorption', []),
        ('nonmem/models/pheno_advan3_trans1.mod', 'absorption', []),
        ('nonmem/models/pheno_noifs.mod', 'absorption', []),
        ('nonmem/models/pheno_trans1.mod', 'absorption', []),
        ('nonmem/secondary_parameters/run1.mod', 'absorption', []),
        ('nonmem/secondary_parameters/run2.mod', 'absorption', []),
        ('nonmem/qa/iov.mod', 'absorption', []),
        ('nonmem/pheno.mod', 'distribution', ['V']),
        ('nonmem/pheno_real.mod', 'distribution', ['V']),
        ('nonmem/models/mox1.mod', 'distribution', ['V']),
        ('nonmem/models/mox2.mod', 'distribution', ['VC']),
        ('nonmem/models/mox_2comp.mod', 'distribution', ['K12', 'K21']),
        ('nonmem/models/pef.mod', 'distribution', []),
        ('nonmem/models/pheno_advan3_trans1.mod', 'distribution', ['K12', 'K21']),
        ('nonmem/models/pheno_noifs.mod', 'distribution', ['V']),
        ('nonmem/models/pheno_trans1.mod', 'distribution', []),
        ('nonmem/secondary_parameters/run1.mod', 'distribution', ['V']),
        ('nonmem/secondary_parameters/run2.mod', 'distribution', ['V']),
        ('nonmem/qa/iov.mod', 'distribution', ['V']),
        ('nonmem/pheno.mod', 'elimination', ['CL']),
        ('nonmem/pheno_real.mod', 'elimination', ['CL']),
        ('nonmem/models/mox1.mod', 'elimination', ['CL']),
        ('nonmem/models/mox2.mod', 'elimination', ['CL']),
        ('nonmem/models/mox_2comp.mod', 'elimination', ['K']),
        ('nonmem/models/pef.mod', 'elimination', ['K']),
        ('nonmem/models/pheno_advan3_trans1.mod', 'elimination', ['K']),
        ('nonmem/models/pheno_noifs.mod', 'elimination', ['CL']),
        ('nonmem/models/pheno_trans1.mod', 'elimination', ['K']),
        ('nonmem/secondary_parameters/run1.mod', 'elimination', ['CL']),
        ('nonmem/secondary_parameters/run2.mod', 'elimination', ['CL']),
        ('nonmem/qa/iov.mod', 'elimination', ['CL']),
        ('nonmem/modeling/pheno_1transit.mod', 'all', ['CL', 'V', 'K12', 'K23', 'K34', 'K43']),
        ('nonmem/modeling/pheno_1transit.mod', 'absorption', ['K12', 'K23']),
        ('nonmem/modeling/pheno_1transit.mod', 'distribution', ['V', 'K34', 'K43']),
        ('nonmem/modeling/pheno_1transit.mod', 'elimination', ['CL']),
        (
            'nonmem/modeling/pheno_2transits.mod',
            'all',
            ['CL', 'V', 'K12', 'K23', 'K34', 'K45', 'K54'],
        ),
        ('nonmem/modeling/pheno_2transits.mod', 'absorption', ['K12', 'K23', 'K34']),
        ('nonmem/modeling/pheno_2transits.mod', 'distribution', ['V', 'K45', 'K54']),
        ('nonmem/modeling/pheno_2transits.mod', 'elimination', ['CL']),
        ('nonmem/modeling/pheno_advan5_nodepot.mod', 'all', ['CL', 'V', 'K12', 'K21']),
        ('nonmem/modeling/pheno_advan5_nodepot.mod', 'absorption', []),
        ('nonmem/modeling/pheno_advan5_nodepot.mod', 'distribution', ['V', 'K12', 'K21']),
        ('nonmem/modeling/pheno_advan5_nodepot.mod', 'elimination', ['CL']),
        ('nonmem/modeling/pheno_advan5_depot.mod', 'all', ['CL', 'V', 'K12', 'K23', 'K32']),
        ('nonmem/modeling/pheno_advan5_depot.mod', 'absorption', ['K12']),
        ('nonmem/modeling/pheno_advan5_depot.mod', 'distribution', ['V', 'K23', 'K32']),
        ('nonmem/modeling/pheno_advan5_depot.mod', 'elimination', ['CL']),
        (
            'nonmem/modeling/transit_indirect_reabsorption.mod',
            'all',
            ['CL', 'V', 'K12', 'K23', 'KA', 'K45', 'K56', 'K64'],
        ),
        (
            'nonmem/modeling/transit_indirect_reabsorption.mod',
            'absorption',
            ['K12', 'K23', 'KA', 'K45'],
        ),
        (
            'nonmem/modeling/transit_indirect_reabsorption.mod',
            'distribution',
            ['V'],
        ),
        ('nonmem/modeling/transit_indirect_reabsorption.mod', 'elimination', ['CL']),
    ),
    ids=repr,
)
def test_get_pk_parameters(load_model_for_test, testdata, model_path, kind, expected):
    model = load_model_for_test(testdata / model_path)
    assert set(get_pk_parameters(model, kind)) == set(expected)

    try:
        pkpd_model = add_effect_compartment(model, "linear")
    except ValueError:  # Model couldn't be transformed
        pass
    else:
        assert set(get_pk_parameters(pkpd_model, kind)) == set(expected)


def test_get_pk_parameters_metabolite(load_model_for_test, testdata):
    model = load_model_for_test(testdata / "nonmem" / "pheno.mod")
    model = add_metabolite(model)

    expected = ['CL', 'CLM', 'V', 'VM']

    assert set(get_pk_parameters(model)) == set(expected)

    model = add_peripheral_compartment(model, "METABOLITE")

    expected.extend(['QP1', 'VP1'])
    assert set(get_pk_parameters(model)) == set(expected)


@pytest.mark.parametrize(
    ('model_path', 'kind', 'expected'),
    (
        # ('nonmem/pheno.mod', 'baseline', ['B']),
        ('nonmem/pheno.mod', 'linear', ['B', 'SLOPE']),
        ('nonmem/pheno.mod', 'emax', ['B', 'E_MAX', 'EC_50']),
        ('nonmem/pheno.mod', 'step', ['B', 'E_MAX']),
        ('nonmem/pheno.mod', 'sigmoid', ['B', 'EC_50', 'E_MAX', 'N']),
    ),
    ids=repr,
)
def test_get_pd_parameters(load_model_for_test, testdata, model_path, kind, expected):
    model = load_model_for_test(testdata / model_path)
    assert set(get_pd_parameters(set_direct_effect(model, kind))) == set(expected)
    assert set(get_pd_parameters(add_effect_compartment(model, kind))) == set(expected + ["MET"])
    assert get_pk_parameters(add_effect_compartment(model, kind)) == ['CL', 'V']
    assert get_pk_parameters(set_direct_effect(model, kind)) == ['CL', 'V']
    assert not set(
        set(get_pd_parameters(set_direct_effect(model, kind))).intersection(
            get_pk_parameters(set_direct_effect(model, kind))
        )
    )
    assert not set(
        set(get_pd_parameters(add_effect_compartment(model, kind))).intersection(
            get_pk_parameters(add_effect_compartment(model, kind))
        )
    )


@pytest.mark.parametrize(
    ('model_path', 'kind', 'prod', 'expected'),
    (
        ('nonmem/pheno.mod', 'linear', True, ['B', 'SLOPE', 'MET']),
        ('nonmem/pheno.mod', 'linear', False, ['B', 'SLOPE', 'MET']),
        ('nonmem/pheno.mod', 'emax', True, ['B', 'E_MAX', 'EC_50', 'MET']),
        ('nonmem/pheno.mod', 'emax', False, ['B', 'E_MAX', 'EC_50', 'MET']),
        ('nonmem/pheno.mod', 'sigmoid', True, ['B', 'EC_50', 'E_MAX', 'N', 'MET']),
        ('nonmem/pheno.mod', 'sigmoid', False, ['B', 'EC_50', 'E_MAX', 'N', 'MET']),
    ),
    ids=repr,
)
def test_get_pd_parameters_indirect(
    load_model_for_test, testdata, model_path, kind, prod, expected
):
    model = load_model_for_test(testdata / model_path)
    assert set(get_pd_parameters(add_indirect_effect(model, kind, prod=prod))) == set(expected)
    assert get_pk_parameters(add_indirect_effect(model, kind, prod=prod)) == ['CL', 'V']
    assert not set(
        set(get_pd_parameters(add_indirect_effect(model, kind, prod=prod))).intersection(
            get_pk_parameters(add_indirect_effect(model, kind, prod=prod))
        )
    )


@pytest.mark.parametrize(
    ('model_path', 'level', 'expected'),
    (
        ('nonmem/pheno.mod', 'all', ['CL', 'V']),
        ('nonmem/pheno.mod', 'random', ['CL', 'V']),
        ('nonmem/pheno.mod', 'iiv', ['CL', 'V']),
        ('nonmem/pheno.mod', 'iov', []),
        ('nonmem/pheno_real.mod', 'all', ['CL', 'V']),
        ('nonmem/pheno_real.mod', 'random', ['CL', 'V']),
        ('nonmem/pheno_real.mod', 'iiv', ['CL', 'V']),
        ('nonmem/pheno_real.mod', 'iov', []),
        ('nonmem/models/mox1.mod', 'all', ['CL', 'KA', 'V']),
        ('nonmem/models/mox1.mod', 'random', ['CL', 'KA', 'V']),
        ('nonmem/models/mox1.mod', 'iiv', ['CL', 'KA', 'V']),
        ('nonmem/models/mox1.mod', 'iov', []),
        ('nonmem/models/mox2.mod', 'all', ['CL', 'MAT', 'VC']),
        ('nonmem/models/mox2.mod', 'random', ['CL', 'MAT', 'VC']),
        ('nonmem/models/mox2.mod', 'iiv', ['CL', 'MAT', 'VC']),
        ('nonmem/models/mox2.mod', 'iov', []),
        ('nonmem/models/mox_2comp.mod', 'all', ['K', 'K12', 'K21']),
        ('nonmem/models/mox_2comp.mod', 'random', ['K', 'K12', 'K21']),
        ('nonmem/models/mox_2comp.mod', 'iiv', ['K', 'K12', 'K21']),
        ('nonmem/models/mox_2comp.mod', 'iov', []),
        ('nonmem/models/pef.mod', 'all', ['K']),
        ('nonmem/models/pef.mod', 'random', ['K']),
        ('nonmem/models/pef.mod', 'iiv', ['K']),
        ('nonmem/models/pef.mod', 'iov', []),
        ('nonmem/models/pheno_advan3_trans1.mod', 'all', ['K', 'K12', 'K21']),
        ('nonmem/models/pheno_advan3_trans1.mod', 'random', ['K', 'K12', 'K21']),
        ('nonmem/models/pheno_advan3_trans1.mod', 'iiv', ['K', 'K12', 'K21']),
        ('nonmem/models/pheno_advan3_trans1.mod', 'iov', []),
        ('nonmem/models/pheno_noifs.mod', 'random', ['CL', 'V']),
        ('nonmem/models/pheno_noifs.mod', 'all', ['CL', 'V']),
        ('nonmem/models/pheno_noifs.mod', 'iiv', ['CL', 'V']),
        ('nonmem/models/pheno_noifs.mod', 'iov', []),
        ('nonmem/models/pheno_trans1.mod', 'all', ['K']),
        ('nonmem/models/pheno_trans1.mod', 'random', ['K']),
        ('nonmem/models/pheno_trans1.mod', 'iiv', ['K']),
        ('nonmem/models/pheno_trans1.mod', 'iov', []),
        ('nonmem/secondary_parameters/run1.mod', 'all', ['CL', 'V']),
        ('nonmem/secondary_parameters/run1.mod', 'random', ['CL', 'V']),
        ('nonmem/secondary_parameters/run1.mod', 'iiv', ['CL', 'V']),
        ('nonmem/secondary_parameters/run1.mod', 'iov', []),
        ('nonmem/secondary_parameters/run2.mod', 'all', ['CL', 'V']),
        ('nonmem/secondary_parameters/run2.mod', 'random', ['CL', 'V']),
        ('nonmem/secondary_parameters/run2.mod', 'iiv', ['CL', 'V']),
        ('nonmem/secondary_parameters/run2.mod', 'iov', []),
        ('nonmem/qa/iov.mod', 'random', ['CL', 'V']),
        ('nonmem/qa/iov.mod', 'random', ['CL', 'V']),
        ('nonmem/qa/iov.mod', 'iiv', ['CL', 'V']),
        ('nonmem/qa/iov.mod', 'iov', ['CL', 'V']),
        ('nonmem/modeling/pheno_1transit.mod', 'all', ['CL', 'V', 'K23', 'K34', 'K43', 'K12']),
        ('nonmem/modeling/pheno_1transit.mod', 'random', ['CL', 'V']),
        ('nonmem/modeling/pheno_1transit.mod', 'iiv', ['CL', 'V']),
        ('nonmem/modeling/pheno_1transit.mod', 'iov', []),
        (
            'nonmem/modeling/pheno_2transits.mod',
            'all',
            ['CL', 'V', 'K34', 'K45', 'K54', 'K12', 'K23'],
        ),
        ('nonmem/modeling/pheno_2transits.mod', 'random', ['CL', 'V']),
        ('nonmem/modeling/pheno_2transits.mod', 'iiv', ['CL', 'V']),
        ('nonmem/modeling/pheno_2transits.mod', 'iov', []),
    ),
    ids=repr,
)
def test_get_individual_parameters(load_model_for_test, testdata, model_path, level, expected):
    model = load_model_for_test(testdata / model_path)
    assert set(get_individual_parameters(model, level)) == set(expected)


basic_pk_model = create_basic_pk_model()


@pytest.mark.parametrize(
    ('func', 'level', 'dv', 'expected'),
    (
        (partial(set_direct_effect, expr='linear'), 'all', None, ['B', 'CL', 'SLOPE', 'VC']),
        (partial(set_direct_effect, expr='linear'), 'all', 1, ['CL', 'VC']),
        (partial(set_direct_effect, expr='linear'), 'all', 2, ['B', 'CL', 'SLOPE', 'VC']),
        (partial(set_direct_effect, expr='emax'), 'all', None, ['B', 'CL', 'EC_50', 'E_MAX', 'VC']),
        (
            partial(set_direct_effect, expr='sigmoid'),
            'all',
            None,
            ['B', 'CL', 'EC_50', 'E_MAX', 'N', 'VC'],
        ),
        (
            partial(add_effect_compartment, expr='sigmoid'),
            'all',
            None,
            ['B', 'CL', 'EC_50', 'E_MAX', 'MET', 'N', 'VC'],
        ),
        (
            partial(add_indirect_effect, expr='linear'),
            'all',
            None,
            ['B', 'CL', 'MET', 'SLOPE', 'VC'],
        ),
        (
            partial(add_indirect_effect, expr='sigmoid'),
            'all',
            None,
            ['B', 'CL', 'EC_50', 'E_MAX', 'MET', 'N', 'VC'],
        ),
    ),
)
def test_get_individual_parameters_pkpd_models(func, level, dv, expected):
    model = func(basic_pk_model)
    params = get_individual_parameters(model, level=level, dv=dv)
    assert params == expected


@pytest.mark.parametrize(
    ('func', 'level', 'dv', 'expected'),
    (
        (add_metabolite, 'all', None, ['CL', 'CLM', 'VC', 'VM']),
        (
            partial(add_metabolite, presystemic=True),
            'all',
            None,
            ['CL', 'CLM', 'FPRE', 'MAT', 'VC', 'VM'],
        ),
    ),
)
def test_get_individual_parameters_drug_metabolite_models(func, level, dv, expected):
    model = func(basic_pk_model)
    params = get_individual_parameters(model, level=level, dv=dv)
    assert params == expected


@pytest.mark.parametrize(
    ('model_path', 'level', 'expected'),
    (
        ('nonmem/pheno.mod', 'all', ['CL', 'V']),
        ('nonmem/pheno_real.mod', 'all', ['CL', 'V']),
        ('nonmem/models/mox1.mod', 'all', ['CL', 'KA', 'V']),
        ('nonmem/models/mox2.mod', 'all', ['CL', 'MAT', 'VC']),
        ('nonmem/models/mox_2comp.mod', 'all', ['K', 'K12', 'K21']),
        ('nonmem/models/pef.mod', 'all', ['K']),
        ('nonmem/models/pheno_advan3_trans1.mod', 'all', ['K', 'K12', 'K21']),
        ('nonmem/models/pheno_noifs.mod', 'all', ['CL', 'V']),
        ('nonmem/models/pheno_trans1.mod', 'all', ['K']),
        ('nonmem/secondary_parameters/run1.mod', 'all', ['CL', 'V']),
        ('nonmem/secondary_parameters/run2.mod', 'all', ['CL', 'V']),
        ('nonmem/qa/iov.mod', 'all', ['CL', 'V']),
        ('nonmem/pheno.mod', 'iiv', ['CL', 'V']),
        ('nonmem/pheno_real.mod', 'iiv', ['CL', 'V']),
        ('nonmem/models/mox1.mod', 'iiv', ['CL', 'KA', 'V']),
        ('nonmem/models/mox2.mod', 'iiv', ['CL', 'MAT', 'VC']),
        ('nonmem/models/mox_2comp.mod', 'iiv', ['K', 'K12', 'K21']),
        ('nonmem/models/pef.mod', 'iiv', ['K']),
        ('nonmem/models/pheno_advan3_trans1.mod', 'iiv', ['K', 'K12', 'K21']),
        ('nonmem/models/pheno_noifs.mod', 'iiv', ['CL', 'V']),
        ('nonmem/models/pheno_trans1.mod', 'iiv', ['K']),
        ('nonmem/secondary_parameters/run1.mod', 'iiv', ['CL', 'V']),
        ('nonmem/secondary_parameters/run2.mod', 'iiv', ['CL', 'V']),
        ('nonmem/qa/iov.mod', 'iiv', ['CL', 'V']),
        ('nonmem/pheno.mod', 'iov', []),
        ('nonmem/pheno_real.mod', 'iov', []),
        ('nonmem/models/mox1.mod', 'iov', []),
        ('nonmem/models/mox2.mod', 'iov', []),
        ('nonmem/models/mox_2comp.mod', 'iov', []),
        ('nonmem/models/pef.mod', 'iov', []),
        ('nonmem/models/pheno_advan3_trans1.mod', 'iov', []),
        ('nonmem/models/pheno_noifs.mod', 'iov', []),
        ('nonmem/models/pheno_trans1.mod', 'iov', []),
        ('nonmem/secondary_parameters/run1.mod', 'iov', []),
        ('nonmem/secondary_parameters/run2.mod', 'iov', []),
        ('nonmem/qa/iov.mod', 'iov', ['CL', 'V']),
        ('nonmem/modeling/pheno_1transit.mod', 'all', ['CL', 'V']),
        ('nonmem/modeling/pheno_1transit.mod', 'iiv', ['CL', 'V']),
        ('nonmem/modeling/pheno_1transit.mod', 'iov', []),
        ('nonmem/modeling/pheno_2transits.mod', 'all', ['CL', 'V']),
        ('nonmem/modeling/pheno_2transits.mod', 'iiv', ['CL', 'V']),
        ('nonmem/modeling/pheno_2transits.mod', 'iov', []),
    ),
    ids=repr,
)
def test_has_random_effect(load_model_for_test, testdata, model_path, level, expected):
    model = load_model_for_test(testdata / model_path)
    params_with_random_effect = set(expected)
    for param in get_pk_parameters(model):
        if param in params_with_random_effect:
            assert has_random_effect(model, param, level)
        else:
            assert not has_random_effect(model, param, level)


@pytest.mark.parametrize(
    ('model_path', 'rv', 'expected'),
    (
        ('nonmem/pheno.mod', 'ETA_1', ['CL']),
        ('nonmem/pheno_real.mod', 'ETA_1', ['CL']),
        ('nonmem/pheno_block.mod', 'ETA_MAT', ['MAT']),
        ('nonmem/qa/iov.mod', 'ETA_1', ['CL']),
        ('nonmem/qa/iov.mod', 'ETA_2', ['V']),
        ('nonmem/qa/iov.mod', 'ETA_3', ['CL']),
        ('nonmem/qa/iov.mod', 'ETA_5', ['CL']),
    ),
    ids=repr,
)
def test_get_rv_parameter(load_model_for_test, testdata, model_path, rv, expected):
    model = load_model_for_test(testdata / model_path)
    rv_params = get_rv_parameters(model, rv)

    assert rv_params == expected


@pytest.mark.parametrize(
    ('model_path', 'param', 'var_type', 'expected'),
    (
        ('nonmem/pheno.mod', 'CL', 'iiv', ['ETA_1']),
        ('nonmem/pheno_real.mod', 'CL', 'iiv', ['ETA_1']),
        ('nonmem/pheno_real.mod', 'TAD', 'iiv', []),
        ('nonmem/qa/iov.mod', 'CL', 'iiv', ['ETA_1']),
        ('nonmem/qa/iov.mod', 'V', 'iiv', ['ETA_2']),
        ('nonmem/qa/iov.mod', 'CL', 'iiv', ['ETA_1']),
        ('nonmem/qa/iov.mod', 'V', 'iiv', ['ETA_2']),
        ('nonmem/qa/iov.mod', 'CL', 'iov', ['ETA_3', 'ETA_5']),
        ('nonmem/qa/iov.mod', 'V', 'iov', ['ETA_4', 'ETA_6']),
    ),
    ids=repr,
)
def test_get_parameter_rv(load_model_for_test, testdata, model_path, param, var_type, expected):
    model = load_model_for_test(testdata / model_path)
    iiv_names = get_parameter_rv(model, param, var_type)
    assert iiv_names == expected


def test_get_rv_parameter_verify_input(load_model_for_test, pheno_path):
    model = load_model_for_test(pheno_path)
    with pytest.raises(ValueError, match='Could not find random variable: x'):
        get_rv_parameters(model, 'x')


def test_get_parameter_rv_verify_input(load_model_for_test, pheno_path):
    model = load_model_for_test(pheno_path)
    with pytest.raises(ValueError, match='Could not find parameter x'):
        get_parameter_rv(model, 'x')
    with pytest.raises(
        ValueError,
        match='ETA_1 is a random variable. Only parameters are accepted as input',
    ):
        get_parameter_rv(model, 'ETA_1')


def test_display_odes(load_model_for_test, pheno_path):
    model = load_model_for_test(pheno_path)
    display_odes(model)


def test_is_real(load_model_for_test, pheno_path):
    model = load_model_for_test(pheno_path)
    assert is_real(model, "CL")
    assert not is_real(model, "I*CL")
    assert is_real(model, "sqrt(CL)") is None


def test_is_linearized(load_example_model_for_test):
    model = load_example_model_for_test("pheno")
    assert not is_linearized(model)
    model = load_example_model_for_test("pheno_linear")
    assert is_linearized(model)


def test_get_dv_symbol(testdata, load_model_for_test):
    model = load_model_for_test(testdata / 'nonmem' / 'models' / 'pheno_dvid.mod')
    assert get_dv_symbol(model, 2) == sympy.Symbol("Y_2")
    assert get_dv_symbol(model, "Y_1") == sympy.Symbol("Y_1")
    assert get_dv_symbol(model, sympy.Symbol("Y_2")) == sympy.Symbol("Y_2")
    assert get_dv_symbol(model) == sympy.Symbol("Y_1")
    assert get_dv_symbol(model, None) == sympy.Symbol("Y_1")

    with pytest.raises(ValueError):
        get_dv_symbol(model, 3)
    with pytest.raises(ValueError):
        get_dv_symbol(model, "FLUMOX")
    with pytest.raises(ValueError):
        get_dv_symbol(model, sympy.Symbol("SPANNER"))
    with pytest.raises(TypeError):
        get_dv_symbol(model, 3.4)



import io
import re
from glob import glob
from os.path import abspath, basename, dirname, join, splitext
from textwrap import dedent

from setuptools import find_packages, setup


def read(*names, **kwargs):
    """Read file path relative to setup.py (in encoding; default: utf8)"""
    return io.open(
        join(dirname(abspath(__file__)), *names), encoding=kwargs.get('encoding', 'utf8')
    ).read()


def strip_refs(text_str):
    """Strip ref text roles (not valid long_description markup)"""
    return re.sub(':[a-z]+:`~?(.*?)`', r'``\1``', text_str)


def longdesc(text_str):
    """Extract blocks between start-longdesc and end-longdesc directives"""
    pat = r'(?<=^\.\. start-longdesc).*?(?=^\.\. end-longdesc)'
    txt = re.compile(pat, re.MULTILINE | re.DOTALL).findall(text_str)
    txt = [dedent(block).strip() for block in txt]
    return '\n\n'.join(txt)


setup(
    name='pharmpy-core',
    version='0.106.0',
    license='GNU Lesser General Public License v3 (LGPLv3)',
    description='Pharmacometric modeling',
    long_description='%s\n\n%s'
    % (strip_refs(longdesc(read('README.rst'))), strip_refs(read('CHANGELOG.rst'))),
    author='Rikard Nordgren',
    author_email='rikard.nordgren@farmaci.uu.se',
    url='https://pharmpy.github.io',
    project_urls={
        "Bug Tracker": 'https://github.com/pharmpy/pharmpy/issues',
        "Source Code": 'https://github.com/pharmpy/pharmpy',
    },
    packages=find_packages('src'),
    package_dir={'': 'src'},
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
    include_package_data=True,
    zip_safe=False,
    python_requires='>=3.10',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Operating System :: POSIX',
        'Operating System :: MacOS',
        'Operating System :: Unix',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Scientific/Engineering',
    ],
    keywords=[
        'pharmacometrics',
    ],
    install_requires=[
        'lark>=1.1.4',
        'sympy>=1.9',
        'symengine>=0.9.2',
        'pandas>=1.4, !=2.1.0',
        'numexpr>=2.7.0',
        'altair>=5.0.0',
        'jsonschema>=3.1.0',
        'sphinx>=2.2.0',
        'csscompressor>=0.9.5',
        'beautifulsoup4>=4.8.0',
        'lxml>=4.4.0',
        'numpy>=1.17',
        'scipy>=1.4.0',
        'dask>=2022.12.1',
        'distributed>=2022.12.1',
        'networkx>=2.3',
        'appdirs>=1.4.4',
        'rich>=4.2.0',
        'jupyter-sphinx>=0.2.0',
        'ipykernel>=5.4.0',
    ],
    extras_require={
        'nlmixr': ["pyreadr"],
    },
    entry_points={
        'console_scripts': [
            'pharmpy             = pharmpy.__main__:run',
            'psn-pharmpy-wrapper = pharmpy.tools.psn_helpers:pharmpy_wrapper',
        ]
    },
)



import pytest


@pytest.mark.parametrize(
    "buf,like,llike",
    [
        ('$ESTIMATION METHOD=1 LIKELIHOOD', True, False),
        ('$ESTIMATION LIKE METHOD=1', True, False),
        ('$ESTIMATION METHOD=1', False, False),
        ('$ESTIMATION METHOD=1 -2LOGLIKELIHOOD', False, True),
        ('$ESTIMATION METHOD=1 -2LL', False, True),
        ('$ESTIMATION METHOD=1 -2LLIKELIHOOD', False, True),
    ],
)
def test_likelihood(parser, buf, like, llike):
    recs = parser.parse(buf)
    rec = recs.records[0]
    assert rec.likelihood == like
    assert rec.loglikelihood == llike



from __future__ import annotations

from abc import abstractmethod
from collections.abc import Collection, Hashable, Sized
from math import sqrt
from typing import TYPE_CHECKING, Any, Dict, List, Mapping, Sequence, Set, Tuple, Union

import pharmpy.internals.unicode as unicode
from pharmpy.internals.expr.parse import parse as parse_expr
from pharmpy.internals.expr.subs import subs
from pharmpy.internals.immutable import Immutable

from .numeric import MultivariateNormalDistribution as NumericMultivariateNormalDistribution
from .numeric import NormalDistribution as NumericNormalDistribution
from .numeric import NumericDistribution

if TYPE_CHECKING:
    import numpy as np
    import symengine
    import sympy
else:
    from pharmpy.deps import numpy as np
    from pharmpy.deps import symengine, sympy


class Distribution(Sized, Hashable, Immutable):
    @abstractmethod
    def replace(self, **kwargs) -> Distribution:
        pass

    @property
    @abstractmethod
    def names(self) -> Tuple[str, ...]:
        """Names of random variables of distribution"""
        pass

    @property
    @abstractmethod
    def level(self) -> str:
        """Name of VariabilityLevel of the random variables"""
        pass

    @property
    @abstractmethod
    def mean(self) -> sympy.Expr:
        pass

    @property
    @abstractmethod
    def variance(self) -> sympy.Expr:
        pass

    @abstractmethod
    def get_variance(self, name: str) -> sympy.Expr:
        pass

    @abstractmethod
    def get_covariance(self, name1: str, name2: str) -> sympy.Expr:
        pass

    @abstractmethod
    def evalf(self, parameters: Dict[sympy.Symbol, float]) -> NumericDistribution:
        pass

    @abstractmethod
    def __getitem__(self, index) -> Distribution:
        pass

    @property
    @abstractmethod
    def free_symbols(self) -> Set[sympy.Expr]:
        pass

    @property
    def parameter_names(self) -> Tuple[str, ...]:
        """List of names of all parameters used in definition"""
        params = self.mean.free_symbols.union(self.variance.free_symbols)
        return tuple(sorted(map(str, params)))

    @abstractmethod
    def subs(self, d: Mapping[Union[str, sympy.Expr], sympy.Expr]) -> Distribution:
        pass

    @abstractmethod
    def latex_string(self, aligned: bool = False) -> str:
        pass

    @abstractmethod
    def to_dict(self) -> Dict[str, Any]:
        pass

    def _repr_latex_(self) -> str:
        return self.latex_string()


class NormalDistribution(Distribution):
    """Normal distribution for one random variable

    Parameters
    ----------
    name : str
        Name of the random variable
    level : str
        Name of the variability level
    mean : expression or number
        Mean of the random variable
    variance : expression or number
        Variance of the random variable

    Example
    -------
    >>> from pharmpy.model import NormalDistribution, Parameter
    >>> omega = Parameter.create('OMEGA_CL', 0.1)
    >>> dist = NormalDistribution.create("IIV_CL", "IIV", 0, omega.symbol)
    >>> dist
    IIV_CL ~ N(0, OMEGA_CL)
    """

    def __init__(self, name: str, level: str, mean: sympy.Expr, variance: sympy.Expr):
        self._name = name
        self._level = level
        self._mean = mean
        self._variance = variance

    @classmethod
    def create(cls, name: str, level: str, mean: sympy.Expr, variance: sympy.Expr):
        level = level.upper()
        mean = parse_expr(mean)
        variance = parse_expr(variance)
        if sympy.ask(sympy.Q.nonnegative(variance)) is False:
            raise ValueError("Variance of normal distribution must be non-negative")
        return cls(name, level, mean, variance)

    def replace(self, **kwargs) -> NormalDistribution:
        """Replace properties and create a new NormalDistribution"""
        name = kwargs.get('name', self._name)
        level = kwargs.get('level', self._level)
        mean = kwargs.get('mean', self._mean)
        variance = kwargs.get('variance', self._variance)
        new = NormalDistribution.create(name, level, mean, variance)
        return new

    @property
    def names(self) -> Tuple[str, ...]:
        return (self._name,)

    @property
    def level(self) -> str:
        return self._level

    @property
    def mean(self) -> sympy.Expr:
        return self._mean

    @property
    def variance(self) -> sympy.Expr:
        return self._variance

    @property
    def free_symbols(self) -> set[sympy.Expr]:
        """Free symbols including random variable itself"""
        fs = self._mean.free_symbols.union(self._variance.free_symbols)
        fs.add(sympy.Symbol(self._name))
        return fs  # pyright: ignore [reportGeneralTypeIssues]

    def subs(self, d: Mapping[Union[str, sympy.Expr], sympy.Expr]) -> NormalDistribution:
        """Substitute expressions

        Parameters
        ----------
        d : dict
            Dictionary of from: to pairs for substitution

        Examples
        --------
        >>> import sympy
        >>> from pharmpy.model import NormalDistribution, Parameter
        >>> omega = Parameter.create("OMEGA_CL", 0.1)
        >>> dist = NormalDistribution.create("IIV_CL", "IIV", 0, omega.symbol)
        >>> dist = dist.subs({omega.symbol: sympy.Symbol("OMEGA_NEW")})
        >>> dist
        IIV_CL ~ N(0, OMEGA_NEW)

        """
        mean = subs(self._mean, d)
        variance = subs(self._variance, d)
        name = _subs_name(self._name, d)
        return NormalDistribution(name, self._level, mean, variance)

    def evalf(self, parameters: Dict[sympy.Symbol, float]) -> NumericDistribution:
        # mu = float(symengine.sympify(rv._mean[0]).xreplace(parameters))
        # sigma = float(symengine.sympify(sympy.sqrt(rv._variance[0,0])).xreplace(parameters))
        mean = self.mean
        assert isinstance(mean, sympy.Expr)
        variance = self.variance
        assert isinstance(variance, sympy.Symbol)
        try:
            if mean == 0:
                mu = 0
            elif isinstance(mean, sympy.Symbol):
                mu = float(parameters[mean])
            else:
                raise NotImplementedError("Non-supported mean for NormalDistribution")
            sigma = 0 if variance == 0 else sqrt(float(parameters[variance]))
            return NumericNormalDistribution(mu, sigma)
        except KeyError as e:
            # NOTE: This handles missing parameter substitutions
            raise ValueError(e)

    def __getitem__(self, index):
        if isinstance(index, int):
            if index != 0:
                raise IndexError(index)

        elif isinstance(index, str):
            if index != self._name:
                raise KeyError(index)

        else:
            if isinstance(index, slice):
                index = list(
                    range(index.start, index.stop, index.step if index.step is not None else 1)
                )
                if index != [0]:
                    raise IndexError(index)

            if isinstance(index, Collection):
                if len(index) != 1 or (self._name not in index and 0 not in index):
                    raise KeyError(index)

            else:
                raise KeyError(index)

        return self

    def get_variance(self, name: str) -> sympy.Expr:
        if name != self._name:
            raise KeyError(name)
        return self._variance

    def get_covariance(self, name1: str, name2: str) -> sympy.Expr:
        if name1 == name2 == self._name:
            return self._variance
        else:
            raise KeyError((name1, name2))

    def __eq__(self, other):
        return (
            isinstance(other, NormalDistribution)
            and self._name == other._name
            and self._level == other._level
            and self._mean == other._mean
            and self._variance == other._variance
        )

    def __len__(self):
        return 1

    def __hash__(self):
        return hash((self._name, self._level, self._mean, self._variance))

    def to_dict(self) -> Dict[str, Any]:
        return {
            'class': self.__class__.__name__,
            'name': self._name,
            'level': self._level,
            'mean': sympy.srepr(self._mean),
            'variance': sympy.srepr(self._variance),
        }

    @classmethod
    def from_dict(cls, d: Mapping[str, Any]):
        return cls(
            name=d['name'],
            level=d['level'],
            mean=sympy.parse_expr(d['mean']),
            variance=sympy.parse_expr(d['variance']),
        )

    def __repr__(self):
        return (
            f'{sympy.pretty(sympy.Symbol(self._name), wrap_line=False, use_unicode=True)}'
            f' ~ {unicode.mathematical_script_capital_n}'
            f'({sympy.pretty(self._mean, wrap_line=False, use_unicode=True)}, '
            f'{sympy.pretty(self._variance, wrap_line=False, use_unicode=True)})'
        )

    def latex_string(self, aligned: bool = False) -> str:
        if aligned:
            align_str = ' & '
        else:
            align_str = ''
        rv = sympy.Symbol(self._name)._repr_latex_()[1:-1]
        mean = self._mean._repr_latex_()[1:-1]
        sigma = (self._variance)._repr_latex_()[1:-1]
        latex = rv + align_str + r'\sim  \mathcal{N} \left(' + mean + ',' + sigma + r'\right)'
        if not aligned:
            latex = '$' + latex + '$'
        return latex


class JointNormalDistribution(Distribution):
    """Joint distribution of random variables

    Parameters
    ----------
    names : list
        Names of the random variables
    level : str
        Variability level
    mean : matrix or list
        Vector of the means of the random variables
    variance : matrix or list of lists
        Covariance matrix of the random variables

    Example
    -------
    >>> from pharmpy.model import JointNormalDistribution, Parameter
    >>> omega_cl = Parameter("OMEGA_CL", 0.1)
    >>> omega_v = Parameter("OMEGA_V", 0.1)
    >>> corr_cl_v = Parameter("OMEGA_CL_V", 0.01)
    >>> dist = JointNormalDistribution.create(["IIV_CL", "IIV_V"], "IIV", [0, 0],
    ...     [[omega_cl.symbol, corr_cl_v.symbol], [corr_cl_v.symbol, omega_v.symbol]])
    >>> dist
    ⎡IIV_CL⎤    ⎧⎡0⎤  ⎡ OMEGA_CL   OMEGA_CL_V⎤⎫
    ⎢      ⎥ ~ N⎪⎢ ⎥, ⎢                      ⎥⎪
    ⎣IIV_V ⎦    ⎩⎣0⎦  ⎣OMEGA_CL_V   OMEGA_V  ⎦⎭

    """

    def __init__(
        self,
        names: Tuple[str, ...],
        level: str,
        mean: sympy.ImmutableMatrix,
        variance: sympy.ImmutableMatrix,
    ):
        self._names = names
        self._level = level
        self._mean = mean
        self._variance = variance
        self._symengine_variance = symengine.Matrix(variance)

    @classmethod
    def create(cls, names: Sequence[str], level: str, mean, variance):
        names = tuple(names)
        level = level.upper()
        mean = sympy.ImmutableMatrix(mean)
        variance = sympy.ImmutableMatrix(variance)
        if variance.is_positive_semidefinite is False:
            raise ValueError(
                'Covariance matrix of joint normal distribution is not positive semidefinite'
            )
        return cls(names, level, mean, variance)

    def replace(self, **kwargs):
        """Replace properties and create a new JointNormalDistribution"""
        names = kwargs.get('names', self._names)
        level = kwargs.get('level', self._level)
        mean = kwargs.get('mean', self._mean)
        variance = kwargs.get('variance', self._variance)
        new = JointNormalDistribution.create(names, level, mean, variance)
        return new

    @property
    def names(self) -> Tuple[str, ...]:
        return self._names

    @property
    def level(self) -> str:
        return self._level

    @property
    def mean(self) -> sympy.ImmutableMatrix:
        return self._mean

    @property
    def variance(self) -> sympy.ImmutableMatrix:
        return self._variance

    @property
    def free_symbols(self) -> Set[sympy.Expr]:
        """Free symbols including random variable itself"""
        return self._mean.free_symbols.union(
            self._variance.free_symbols, (sympy.Symbol(name) for name in self._names)
        )

    def subs(self, d: Mapping[Union[str, sympy.Expr], sympy.Expr]) -> JointNormalDistribution:
        """Substitute expressions

        Parameters
        ----------
        d : dict
            Dictionary of from: to pairs for substitution

        Examples
        --------
        >>> import sympy
        >>> from pharmpy.model import JointNormalDistribution, Parameter
        >>> omega1 = Parameter("OMEGA_CL", 0.1)
        >>> omega2 = Parameter("OMEGA_V", 0.1)
        >>> A = [[omega1.symbol, 0], [0, omega2.symbol]]
        >>> dist = JointNormalDistribution.create(["IIV_CL", "IIV_V"], "IIV", [0, 0], A)
        >>> dist = dist.subs({omega1.symbol: sympy.Symbol("OMEGA_NEW")})
        >>> dist
                ⎡IIV_CL⎤    ⎧⎡0⎤  ⎡OMEGA_NEW     0   ⎤⎫
                ⎢      ⎥ ~ N⎪⎢ ⎥, ⎢                  ⎥⎪
                ⎣IIV_V ⎦    ⎩⎣0⎦  ⎣    0      OMEGA_V⎦⎭

        """
        mean = self._mean.subs(d)
        variance = self._variance.subs(d)
        new_names = tuple(_subs_name(name, d) for name in self._names)
        return JointNormalDistribution(new_names, self._level, mean, variance)

    def evalf(self, parameters: Dict[sympy.Symbol, float]) -> NumericDistribution:
        try:
            mu = np.array(symengine.sympify(self.mean).xreplace(parameters)).astype(np.float64)[
                :, 0
            ]
            sigma = np.array(self._symengine_variance.xreplace(parameters)).astype(np.float64)
            return NumericMultivariateNormalDistribution(mu, sigma)
        except RuntimeError as e:
            # NOTE: This handles missing parameter substitutions
            raise ValueError(e)

    def __getitem__(self, index):
        if isinstance(index, int):
            if -len(self) <= index < len(self):
                names = (self._names[index],)
            else:
                raise IndexError(index)

        elif isinstance(index, str):
            names = (index,)
            try:
                index = self._names.index(index)
            except ValueError:
                raise KeyError(names[0])

        else:
            if isinstance(index, slice):
                index = range(index.start, index.stop, index.step if index.step is not None else 1)
                index = [self._names[i] for i in index]

            if isinstance(index, Collection):
                if len(index) == 0 or len(index) > len(self._names):
                    raise KeyError(index)

                collection = set(index)
                if not collection.issubset(self._names):
                    raise KeyError(index)

                if len(collection) == len(self._names):
                    return self

                index_list: List[int] = []
                names_list: List[str] = []

                for i, name in enumerate(self._names):
                    if name in collection:
                        index_list.append(i)
                        names_list.append(name)

                index = tuple(index_list)
                names = tuple(names_list)

                if len(index) == 1:
                    index = index[0]

            else:
                raise KeyError(index)

        mean = self._mean[index, [0]] if isinstance(index, int) else self._mean[index, :]
        variance = self._variance[index, index]

        if len(names) == 1:
            assert isinstance(mean, sympy.Expr)
            assert isinstance(variance, sympy.Expr)
            return NormalDistribution(names[0], self._level, mean, variance)
        else:
            assert isinstance(mean, sympy.ImmutableMatrix)
            assert isinstance(variance, sympy.ImmutableMatrix)
            return JointNormalDistribution(names, self._level, mean, variance)

    def get_variance(self, name: str) -> sympy.Expr:
        i = self._names.index(name)
        var = self._variance[i, i]
        assert isinstance(var, sympy.Expr)
        return var

    def get_covariance(self, name1: str, name2: str) -> sympy.Expr:
        i1 = self._names.index(name1)
        i2 = self._names.index(name2)
        cov = self._variance[i1, i2]
        assert isinstance(cov, sympy.Expr)
        return cov

    def __eq__(self, other):
        return (
            isinstance(other, JointNormalDistribution)
            and self._names == other._names
            and self._level == other._level
            and self._mean == other._mean
            and self._variance == other._variance
        )

    def __len__(self):
        return len(self._names)

    def __hash__(self):
        return hash((self._names, self._level, self._mean, self._variance))

    def to_dict(self) -> Dict[str, Any]:
        return {
            'class': self.__class__.__name__,
            'names': self._names,
            'level': self._level,
            'mean': sympy.srepr(self._mean),
            'variance': sympy.srepr(self._variance),
        }

    @classmethod
    def from_dict(cls, d: Mapping[str, Any]):
        return cls(
            names=d['names'],
            level=d['level'],
            mean=sympy.parse_expr(d['mean']),
            variance=sympy.parse_expr(d['variance']),
        )

    def __repr__(self):
        name_vector = sympy.Matrix(self._names)
        name_strings = sympy.pretty(name_vector, wrap_line=False, use_unicode=True).split('\n')
        mu_strings = sympy.pretty(self._mean, wrap_line=False, use_unicode=True).split('\n')
        sigma_strings = sympy.pretty(self._variance, wrap_line=False, use_unicode=True).split('\n')
        mu_height = len(mu_strings)
        sigma_height = len(sigma_strings)
        max_height = max(mu_height, sigma_height)

        left_parens = unicode.left_parens(max_height)
        right_parens = unicode.right_parens(max_height)

        # Pad the smaller of the matrices
        if mu_height != sigma_height:
            to_pad = mu_strings if mu_height < sigma_height else sigma_strings
            num_lines = abs(mu_height - sigma_height)
            padding = ' ' * len(to_pad[0])
            for i in range(0, num_lines):
                if i % 2 == 0:
                    to_pad.append(padding)
                else:
                    to_pad.insert(0, padding)

        # Pad names
        if len(name_strings) < max_height:
            num_lines = abs(max_height - len(name_strings))
            padding = ' ' * len(name_strings[0])
            for i in range(0, num_lines):
                if i % 2 == 0:
                    name_strings.append(padding)
                else:
                    name_strings.insert(0, padding)

        central_index = max_height // 2
        res = []
        enumerator = enumerate(
            zip(name_strings, left_parens, mu_strings, sigma_strings, right_parens)
        )
        for i, (name_line, lpar, mu_line, sigma_line, rpar) in enumerator:
            if i == central_index:
                res.append(
                    name_line
                    + f' ~ {unicode.mathematical_script_capital_n}'
                    + lpar
                    + mu_line
                    + ', '
                    + sigma_line
                    + rpar
                )
            else:
                res.append(name_line + '    ' + lpar + mu_line + '  ' + sigma_line + rpar)
        return '\n'.join(res)

    def latex_string(self, aligned: bool = False) -> str:
        if aligned:
            align_str = ' & '
        else:
            align_str = ''
        names = [sympy.Symbol(name) for name in self._names]
        rv_vec = sympy.Matrix(names)._repr_latex_()[1:-1]
        mean_vec = self._mean._repr_latex_()[1:-1]
        sigma = self._variance._repr_latex_()[1:-1]
        latex = (
            rv_vec + align_str + r'\sim \mathcal{N} \left(' + mean_vec + ',' + sigma + r'\right)'
        )
        if not aligned:
            latex = '$' + latex + '$'
        return latex

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['_symengine_variance']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._symengine_variance = symengine.sympify(self._variance)


def _subs_name(name: str, d: Mapping[Union[str, sympy.Expr], sympy.Expr]) -> str:
    if name in d:
        new_name = d[name]
    elif (name_symbol := sympy.Symbol(name)) in d:
        new_name = d[name_symbol]
    else:
        new_name = name
    return new_name if isinstance(new_name, str) else str(new_name)



import pytest


@pytest.fixture(scope='session')
def pheno_ext(datadir):
    return datadir / 'pheno_real.ext'


@pytest.fixture(scope='session')
def pheno_phi(datadir):
    return datadir / 'pheno_real.phi'


@pytest.fixture(scope='session')
def pheno_cov(datadir):
    return datadir / 'pheno_real.cov'


@pytest.fixture(scope='session')
def pheno_lst(datadir):
    return datadir / 'pheno_real.lst'


@pytest.fixture(scope='session')
def pheno_data(datadir):
    return datadir / 'pheno.dta'



from pharmpy.tools.simeval.results import psn_simeval_results


def test_psn_simeval_results(testdata):
    res = psn_simeval_results(testdata / 'psn' / 'simeval_dir1')
    assert len(res.sampled_iofv) == 59
    assert len(res.iofv_summary) == 59



from pharmpy.modeling import read_model


def test_dataset(load_model_for_test, testdata):
    model = read_model(testdata / 'nonmem' / 'fcon' / 'FCON')
    assert model.internals.code.startswith('FILE')

    df = model.dataset

    nmtran = load_model_for_test(testdata / 'nonmem' / 'pheno_real.mod')

    assert list(df['WGT']) == list(nmtran.dataset['WGT'])



from pathlib import Path

from pharmpy.internals.fs.tmp import TemporaryDirectory


def test_tempdir():
    with TemporaryDirectory() as td:
        path = Path(td)
        assert path.is_dir()
    assert not path.is_dir()



from dataclasses import dataclass
from itertools import product
from typing import Literal, Optional, Tuple, Union

from lark.visitors import Interpreter

from pharmpy.model import Model
from pharmpy.modeling import get_pk_parameters

from .feature import ModelFeature, feature
from .symbols import Option, Symbol, Wildcard


@dataclass(frozen=True)
class Covariate(ModelFeature):
    parameter: Union[Symbol, Tuple[str, ...]]
    covariate: Union[Symbol, Tuple[str, ...]]
    fp: Tuple[str, ...]
    op: Literal['*', '+'] = '*'
    optional: Option = Option(False)

    def eval(self, model: Optional[Model] = None, explicit_covariates: Optional[set] = None):
        # Circular import issue
        from ...feature.covariate import _interpret_ref

        if model is not None:
            if self.parameter is ParameterWildcard:
                parameter = tuple(get_pk_parameters(model))
            elif isinstance(self.parameter, Ref):
                parameter = _interpret_ref(model, self.parameter)
            else:
                parameter = self.parameter

            if self.covariate is CovariateWildcard:
                covariate = tuple(
                    column.name for column in model.datainfo if column.type == 'covariate'
                )
            elif isinstance(self.covariate, Ref):
                covariate = _interpret_ref(model, self.covariate)
            else:
                covariate = self.covariate
        else:
            parameter = self.parameter
            covariate = self.covariate

        if self.fp == EffectFunctionWildcard:
            fp = (
                'lin',
                'piece_lin',
                'exp',
                'pow',
            )
        else:
            fp = self.fp

        op = self.op

        optional = self.optional

        if explicit_covariates and isinstance(self.parameter, Ref):
            param_cov = set(product(parameter, covariate))
            explicit_to_remove = [
                p for p, c in param_cov.intersection(explicit_covariates) if c in covariate
            ]
            parameter = tuple([p for p in parameter if p not in explicit_to_remove])
        if len(parameter) == 0 or len(covariate) == 0:
            return None
        return Covariate(parameter=parameter, covariate=covariate, fp=fp, op=op, optional=optional)


@dataclass(frozen=True)
class Ref(Symbol):
    name: str


ParameterWildcard = Wildcard()
CovariateWildcard = Wildcard()
EffectFunctionWildcard = Wildcard()


class CovariateInterpreter(Interpreter):
    def interpret(self, tree):
        children = self.visit_children(tree)
        assert 3 <= len(children) <= 5
        if isinstance(children[0], Option):
            if len(children) == 4:
                children.append('*')
            children.append(children.pop(0))
        return feature(Covariate, children)

    def option(self, tree):
        children = self.visit_children(tree)
        assert len(children) == 1
        child = children[0]
        return child if isinstance(child, (Symbol, list)) else [child]

    def parameter_option(self, tree):
        return self.option(tree)

    def covariate_option(self, tree):
        return self.option(tree)

    def fp_option(self, tree):
        return self.option(tree)

    def optional_cov(self, tree):
        children = self.visit_children(tree)
        assert len(children) == 1
        value = children[0].value
        assert isinstance(value, str)
        return Option(True)

    def op_option(self, tree):
        children = self.visit_children(tree)
        assert len(children) == 1
        value = children[0].value
        assert isinstance(value, str)
        return value

    def value(self, tree):
        children = self.visit_children(tree)
        assert len(children) == 1
        value = children[0].value
        assert isinstance(value, str)
        return value.upper()

    def parameter_wildcard(self, tree):
        return ParameterWildcard

    def covariate_wildcard(self, tree):
        return CovariateWildcard

    def fp_wildcard(self, tree):
        return EffectFunctionWildcard

    def ref(self, tree):
        children = self.visit_children(tree)
        assert len(children) == 1
        name = children[0].value
        assert isinstance(name, str)
        return Ref(name)



from pathlib import Path

import pytest
from numpy import nan

import pharmpy.model.external.nonmem.table as table
import pharmpy.tools.external.nonmem.results_file as rf
from pharmpy.tools import read_modelfit_results
from pharmpy.workflows.log import Log


def test_supported_version():
    assert rf.NONMEMResultsFile.supported_version(None) is False
    assert rf.NONMEMResultsFile.supported_version('7.1.0') is False
    assert rf.NONMEMResultsFile.supported_version('7.2.0') is True
    assert rf.NONMEMResultsFile.supported_version('7.3.0') is True


def test_data_io(pheno_lst):
    rfile = rf.NONMEMResultsFile(pheno_lst)
    assert rfile.nonmem_version == "7.4.2"


@pytest.mark.parametrize(
    'file, table_number, expected, covariance_step_ok',
    [
        (
            'phenocorr.lst',
            1,
            {
                'minimization_successful': True,
                'estimate_near_boundary': False,
                'rounding_errors': False,
                'maxevals_exceeded': False,
                'significant_digits': 4.9,
                'function_evaluations': 98,
                'warning': False,
            },
            True,
        ),
        (
            'hessian_error.lst',
            1,
            {
                'minimization_successful': False,
                'estimate_near_boundary': None,
                'rounding_errors': None,
                'maxevals_exceeded': None,
                'significant_digits': nan,
                'function_evaluations': nan,
                'warning': None,
            },
            False,
        ),
        (
            'large_s_matrix_cov_fail.lst',
            1,
            {
                'minimization_successful': True,
                'estimate_near_boundary': False,
                'rounding_errors': False,
                'maxevals_exceeded': False,
                'significant_digits': 3.1,
                'function_evaluations': 62,
                'warning': True,
            },
            False,
        ),
        (
            'nm710_fail_negV.lst',
            1,
            {
                'minimization_successful': None,
                'estimate_near_boundary': None,
                'rounding_errors': None,
                'maxevals_exceeded': None,
                'significant_digits': nan,
                'function_evaluations': nan,
                'warning': None,
            },
            None,
        ),
        (
            'sparse_matrix_with_msfi.lst',
            1,
            {
                'minimization_successful': True,
                'estimate_near_boundary': False,
                'rounding_errors': False,
                'maxevals_exceeded': False,
                'significant_digits': 3.1,
                'function_evaluations': 112,
                'warning': True,
            },
            True,
        ),
        (
            'warfarin_ddmore.lst',
            1,
            {
                'minimization_successful': True,
                'estimate_near_boundary': False,
                'rounding_errors': False,
                'maxevals_exceeded': False,
                'significant_digits': nan,
                'function_evaluations': nan,
                'warning': False,
            },
            False,
        ),
        (
            'mox_fail_nonp.lst',
            1,
            {
                'minimization_successful': False,
                'estimate_near_boundary': False,
                'rounding_errors': False,
                'maxevals_exceeded': False,
                'significant_digits': nan,
                'function_evaluations': 153,
                'warning': False,
            },
            False,
        ),
        (
            'mox_nocov_nonp.lst',
            1,
            {
                'minimization_successful': False,
                'estimate_near_boundary': False,
                'rounding_errors': False,
                'maxevals_exceeded': False,
                'significant_digits': nan,
                'function_evaluations': 153,
                'warning': False,
            },
            False,
        ),
        (
            'pheno_nonp.lst',
            1,
            {
                'minimization_successful': True,
                'estimate_near_boundary': False,
                'rounding_errors': False,
                'maxevals_exceeded': False,
                'significant_digits': 3.6,
                'function_evaluations': 107,
                'warning': False,
            },
            True,
        ),
        (
            'theo.lst',
            1,
            {
                'minimization_successful': True,
                'estimate_near_boundary': False,
                'rounding_errors': False,
                'maxevals_exceeded': False,
                'significant_digits': 4.2,
                'function_evaluations': 208,
                'warning': False,
            },
            True,
        ),
        (
            'theo_nonp.lst',
            1,
            {
                'minimization_successful': False,
                'estimate_near_boundary': True,
                'rounding_errors': True,
                'maxevals_exceeded': False,
                'significant_digits': nan,
                'function_evaluations': 735,
                'warning': False,
            },
            False,
        ),
        (
            'theo_withcov.lst',
            1,
            {
                'minimization_successful': True,
                'estimate_near_boundary': False,
                'rounding_errors': False,
                'maxevals_exceeded': False,
                'significant_digits': 4.2,
                'function_evaluations': 208,
                'warning': False,
            },
            True,
        ),
        (
            'UseCase7.lst',
            1,
            {
                'minimization_successful': True,
                'estimate_near_boundary': False,
                'rounding_errors': False,
                'maxevals_exceeded': False,
                'significant_digits': nan,
                'function_evaluations': nan,
                'warning': False,
            },
            False,
        ),
        (
            'example6b_V7_30_beta.lst',
            1,
            {
                'minimization_successful': True,
                'estimate_near_boundary': False,
                'rounding_errors': False,
                'maxevals_exceeded': False,
                'significant_digits': nan,
                'function_evaluations': nan,
                'warning': False,
            },
            False,
        ),
        (
            'maxeval3.lst',
            1,
            {
                'minimization_successful': False,
                'estimate_near_boundary': False,
                'rounding_errors': False,
                'maxevals_exceeded': True,
                'significant_digits': nan,
                'function_evaluations': 5,
                'warning': False,
            },
            False,
        ),
    ],
)
def test_estimation_status(testdata, file, table_number, expected, covariance_step_ok):
    p = Path(testdata / 'nonmem' / 'modelfit_results' / 'onePROB' / 'oneEST' / 'noSIM')
    log = Log()
    rfile = rf.NONMEMResultsFile(p / file, log=log)
    assert rfile.estimation_status(table_number) == expected
    if covariance_step_ok is None:
        assert rfile.covariance_status(table_number)['covariance_step_ok'] is None
    else:
        assert rfile.covariance_status(table_number)['covariance_step_ok'] == covariance_step_ok


@pytest.mark.parametrize(
    'file, table_number, expected, covariance_step_ok',
    [
        (
            'anneal2_V7_30_beta.lst',
            2,
            {
                'minimization_successful': True,
                'estimate_near_boundary': False,
                'rounding_errors': False,
                'maxevals_exceeded': False,
                'significant_digits': nan,
                'function_evaluations': nan,
                'warning': False,
            },
            True,
        ),
        (
            'superid2_6_V7_30_beta.lst',
            2,
            {
                'minimization_successful': True,
                'estimate_near_boundary': False,
                'rounding_errors': False,
                'maxevals_exceeded': False,
                'significant_digits': nan,
                'function_evaluations': nan,
                'warning': False,
            },
            True,
        ),
    ],
)
def test_estimation_status_multest(testdata, file, table_number, expected, covariance_step_ok):
    p = Path(testdata / 'nonmem' / 'modelfit_results' / 'onePROB' / 'multEST' / 'noSIM')
    rfile = rf.NONMEMResultsFile(p / file)
    assert rfile.estimation_status(table_number) == expected
    assert rfile.covariance_status(table_number)['covariance_step_ok'] == covariance_step_ok


def test_estimation_status_empty():
    rfile = rf.NONMEMResultsFile()
    assert rfile._supported_nonmem_version is False
    assert rfile.estimation_status(1) == rf.NONMEMResultsFile.unknown_termination()


def test_estimation_status_withsim(testdata):
    p = Path(testdata / 'nonmem' / 'modelfit_results' / 'onePROB' / 'oneEST' / 'withSIM')
    rfile = rf.NONMEMResultsFile(p / 'control3boot.res', log=Log())

    assert rfile.estimation_status(45) == {
        'minimization_successful': True,
        'estimate_near_boundary': False,
        'rounding_errors': False,
        'maxevals_exceeded': False,
        'significant_digits': 3.3,
        'function_evaluations': 192,
        'warning': False,
    }
    assert rfile.covariance_status(45)['covariance_step_ok'] is False

    assert rfile.estimation_status(70) == {
        'minimization_successful': True,
        'estimate_near_boundary': True,
        'rounding_errors': False,
        'maxevals_exceeded': False,
        'significant_digits': 3.6,
        'function_evaluations': 202,
        'warning': False,
    }
    assert rfile.covariance_status(70)['covariance_step_ok'] is False

    assert rfile.estimation_status(100) == {
        'minimization_successful': True,
        'estimate_near_boundary': False,
        'rounding_errors': False,
        'maxevals_exceeded': False,
        'significant_digits': 5.6,
        'function_evaluations': 100,
        'warning': False,
    }
    assert rfile.covariance_status(100)['covariance_step_ok'] is True


def test_ofv_table_gap(testdata):
    p = Path(testdata / 'nonmem' / 'modelfit_results' / 'multPROB' / 'multEST' / 'withSIM')
    rfile = rf.NONMEMResultsFile(p / 'multprobmix_nm730.lst', log=Log())

    assert rfile.estimation_status(2) == {
        'minimization_successful': False,
        'estimate_near_boundary': False,
        'rounding_errors': False,
        'maxevals_exceeded': True,
        'significant_digits': nan,
        'function_evaluations': 16,
        'warning': False,
    }

    table_numbers = (1, 2, 3, 4, 6, 8, 10, 11, 12, 13)
    ext_table_file = table.NONMEMTableFile(p / 'multprobmix_nm730.ext')

    for n in table_numbers:
        assert rfile.ofv(n) == pytest.approx(ext_table_file.table_no(n).final_ofv)


@pytest.mark.parametrize(
    'file_name, ref_start, no_of_rows, idx, no_of_errors',
    [
        (
            'control_stream_error.lst',
            'AN ERROR WAS FOUND IN THE CONTROL STATEMENTS.',
            6,
            0,
            1,
        ),
        (
            'no_header_error.lst',
            'PRED EXIT CODE = 1',
            9,
            1,
            2,
        ),
        (
            'no_header_error.lst',
            'PROGRAM TERMINATED BY OBJ',
            2,
            2,
            2,
        ),
        (
            'rounding_error.lst',
            'MINIMIZATION TERMINATED\nDUE TO ROUNDING',
            2,
            0,
            2,
        ),
        (
            'zero_gradient_error.lst',
            'MINIMIZATION TERMINATED\nDUE TO ZERO',
            2,
            0,
            2,
        ),
    ],
)
def test_errors(testdata, file_name, ref_start, no_of_rows, idx, no_of_errors):
    lst = rf.NONMEMResultsFile(testdata / 'nonmem' / 'errors' / file_name, log=Log())
    log_df = lst.log.to_dataframe()
    message = log_df['message'].iloc[idx]
    assert message.startswith(ref_start)
    assert len(message.split('\n')) == no_of_rows
    assert log_df['category'].value_counts()['ERROR'] == no_of_errors


@pytest.mark.parametrize(
    'file_name, ref, idx',
    [
        (
            'no_header_error.lst',
            'THE NUMBER OF PARAMETERS TO BE ESTIMATED\n'
            'EXCEEDS THE NUMBER OF INDIVIDUALS WITH DATA.',
            0,
        ),
        (
            'estimate_near_boundary_warning.lst',
            'PARAMETER ESTIMATE IS NEAR ITS BOUNDARY',
            0,
        ),
        (
            'est_step_warning.lst',
            'MINIMIZATION SUCCESSFUL\nHOWEVER, PROBLEMS OCCURRED WITH THE MINIMIZATION.',
            0,
        ),
    ],
)
def test_warnings(testdata, file_name, ref, idx):
    lst = rf.NONMEMResultsFile(testdata / 'nonmem' / 'errors' / file_name, log=Log())
    message = lst.log.to_dataframe()['message'].iloc[idx]
    assert message == ref


def test_covariance_status(testdata):
    res = read_modelfit_results(
        testdata / 'nonmem' / 'modelfit_results' / 'covariance' / 'pheno_nocovariance.mod'
    )
    assert all(res.standard_errors.isna())
    assert res.covariance_matrix is None
    assert res.correlation_matrix is None
    assert res.precision_matrix is None



import shutil
import tempfile
from pathlib import Path

import pytest

from pharmpy.internals.fs.cwd import chdir
from pharmpy.model import Model
from pharmpy.tools import fit


@pytest.fixture(scope='session')
def start_modelres(testdata):
    tempdir = Path(tempfile.mkdtemp())
    with chdir(tempdir):
        shutil.copy2(testdata / 'nonmem' / 'models' / 'mox2.mod', tempdir)
        shutil.copy2(testdata / 'nonmem' / 'models' / 'mox_simulated_normal.csv', tempdir)
        model_start = Model.parse_model('mox2.mod')
        model_start = model_start.replace(
            datainfo=model_start.datainfo.replace(path=tempdir / 'mox_simulated_normal.csv')
        )
        modelfit_results = fit(model_start)
    return model_start, modelfit_results


@pytest.fixture(scope='session')
def model_count():
    def _model_count(rundir: Path):
        return sum(
            map(
                lambda path: 0 if path.name in ['.lock', '.datasets', 'input_model'] else 1,
                ((rundir / 'models').iterdir()),
            )
        )

    return _model_count



import pytest

from pharmpy.deps import sympy
from pharmpy.model import (
    Assignment,
    Compartment,
    CompartmentalSystem,
    CompartmentalSystemBuilder,
    output,
)
from pharmpy.modeling import (
    add_effect_compartment,
    add_indirect_effect,
    set_baseline_effect,
    set_direct_effect,
    set_michaelis_menten_elimination,
)


def S(x):
    return sympy.Symbol(x)


@pytest.mark.parametrize(
    'pd_model',
    [('linear'), ('emax'), ('sigmoid'), ('step'), ('loglin')],
)
def test_set_direct_effect(load_model_for_test, pd_model, testdata):
    model = load_model_for_test(testdata / "nonmem" / "pheno_pd.mod")
    conc = model.statements.ode_system.central_compartment.amount / S("VC")
    _test_effect_models(set_direct_effect(model, pd_model), pd_model, conc)


@pytest.mark.parametrize(
    'pd_model',
    [('linear'), ('emax'), ('sigmoid'), ('step'), ('loglin')],
)
def test_add_effect_compartment(load_model_for_test, pd_model, testdata):
    model = load_model_for_test(testdata / "nonmem" / "pheno_pd.mod")
    conc_e = sympy.Function("A_EFFECT")('t')
    ke0 = S("KE0")
    central_amount = sympy.Function("A_CENTRAL")(S('t'))
    comp_e = Compartment.create("EFFECT", input=ke0 * central_amount / S("VC"))

    model1 = add_effect_compartment(model, "linear")
    compartments = CompartmentalSystemBuilder(model1.statements.ode_system)
    odes = model1.statements.ode_system
    assert odes.find_compartment("EFFECT") == comp_e
    assert odes.zero_order_inputs[0] == 0
    assert odes.zero_order_inputs[1] == ke0 * central_amount / S("VC")
    assert odes.get_compartment_outflows("EFFECT")[0][0] == output
    assert odes.get_compartment_outflows("EFFECT")[0][1] == ke0
    assert CompartmentalSystem(compartments).compartment_names == ['CENTRAL', 'EFFECT']

    _test_effect_models(add_effect_compartment(model, pd_model), pd_model, conc_e)


def _test_effect_models(model, expr, conc):
    e = S("E")
    e0 = S("B")
    emax = S("E_MAX")
    ec50 = S("EC_50")

    if expr == 'linear':
        assert model.statements[1] == Assignment(e0, S("POP_B"))
        assert model.statements[0] == Assignment(S("SLOPE"), S("POP_SLOPE"))
        assert model.statements.after_odes[-2] == Assignment(e, e0 * (1 + S("SLOPE") * conc))
        assert model.statements.after_odes[-1] == Assignment(S("Y_2"), e + e * S("epsilon_p"))
    elif expr == "emax":
        assert model.statements[0] == Assignment(ec50, S("POP_EC_50"))
        assert model.statements[2] == Assignment(e0, S("POP_B"))
        assert model.statements[1] == Assignment(emax, S("POP_E_MAX"))
        assert model.statements.after_odes[-2] == Assignment(
            e, e0 * (1 + (emax * conc) / (ec50 + conc))
        )
        assert model.statements.after_odes[-1] == Assignment(S("Y_2"), e + e * S("epsilon_p"))
    elif expr == "sigmoid":
        assert model.statements[0] == Assignment(S("N"), S("POP_N"))
        assert model.statements[1] == Assignment(ec50, S("POP_EC_50"))
        assert model.statements[3] == Assignment(e0, S("POP_B"))
        assert model.statements[2] == Assignment(emax, S("POP_E_MAX"))
        assert model.statements.after_odes[-2] == Assignment(
            e,
            sympy.Piecewise(
                (
                    e0 * (1 + ((emax * conc ** S("N")) / (ec50 ** S("N") + conc ** S("N")))),
                    conc > 0,
                ),
                (e0, True),
            ),
        )
        assert model.statements.after_odes[-1] == Assignment(S("Y_2"), e + e * S("epsilon_p"))
        assert model.parameters["POP_N"].init == 1
    elif expr == "step":
        assert model.statements[1] == Assignment(e0, S("POP_B"))
        assert model.statements[0] == Assignment(emax, S("POP_E_MAX"))
        assert model.statements.after_odes[-2] == Assignment(
            e, sympy.Piecewise((e0, conc <= 0), (e0 * (1 + emax), True))
        )
        assert model.statements.after_odes[-1] == Assignment(S("Y_2"), e + e * S("epsilon_p"))
    elif expr == "loglin":
        assert model.statements[1] == Assignment(e0, S("POP_B"))
        assert model.statements[0] == Assignment(S("SLOPE"), S("POP_SLOPE"))
        assert model.statements.after_odes[-2] == Assignment(
            e, S("SLOPE") * sympy.log(conc + sympy.exp(e0 / S("SLOPE")))
        )
        assert model.statements.after_odes[-1] == Assignment(S("Y_2"), e + e * S("epsilon_p"))


@pytest.mark.parametrize(
    'prod, expr',
    [
        (True, 'linear'),
        (True, 'emax'),
        (True, 'sigmoid'),
        (False, 'linear'),
        (False, 'emax'),
        (False, 'sigmoid'),
    ],
)
def test_indirect_effect(load_model_for_test, testdata, prod, expr):
    model = load_model_for_test(testdata / "nonmem" / "pheno_pd.mod")
    model = add_indirect_effect(
        model,
        prod=prod,
        expr=expr,
    )


def test_set_baseline_effect(load_model_for_test, testdata):
    model = load_model_for_test(testdata / "nonmem" / "pheno_pd.mod")
    baseline = set_baseline_effect(model)

    e, e0 = S('E'), S('B')
    assert baseline.statements[0] == Assignment(e0, S("POP_B"))
    assert baseline.statements.after_odes[-2] == Assignment(e, e0)
    assert baseline.statements.after_odes[-1] == Assignment(S("Y_2"), e + e * S("epsilon_p"))


def test_pd_michaelis_menten(load_model_for_test, testdata):
    model = load_model_for_test(testdata / "nonmem" / "pheno_pd.mod")
    pkpd = set_direct_effect(model, 'linear')
    michaelis_menten = set_michaelis_menten_elimination(model)
    pkpd_mm = set_direct_effect(michaelis_menten, 'linear')
    assert pkpd.statements.find_assignment('E') == pkpd_mm.statements.find_assignment('E')

    pkpd = add_effect_compartment(model, 'linear')
    michaelis_menten = set_michaelis_menten_elimination(model)
    pkpd_mm = add_effect_compartment(michaelis_menten, 'linear')
    assert pkpd.statements.find_assignment('E') == pkpd_mm.statements.find_assignment('E')



import itertools
import warnings
from pathlib import Path

from pharmpy.internals.ds.ordered_set import OrderedSet
from pharmpy.model import Model
from pharmpy.modeling import (
    get_covariate_baselines,
    list_time_varying_covariates,
    set_covariates,
    write_model,
)
from pharmpy.tools import read_modelfit_results

from .models import create_model3b


def setup(model_path, covariates):
    input_model = Model.parse_model(model_path)
    covariates = check_covariates(input_model, covariates)
    return covariates


def check_covariates(input_model, covariates):
    """Perform checks of covariates and filter out inappropriate ones

    1. A covariate has the same baseline value for all individuals
    2. Two or more covariates have the exact same baseline values
    3. Warn for time varying covariates

    Return a new list of covariates with "bad" ones removed
    """
    kept = []
    for cov in covariates:
        if cov not in input_model.datainfo.names:
            warnings.warn(f'The covariate {cov} is not available in the dataset. Will be skipped.')
        else:
            kept.append(cov)
    covariates = kept

    input_model = set_covariates(input_model, covariates)
    cov_bls = get_covariate_baselines(input_model)

    tvar = list_time_varying_covariates(input_model)
    if tvar:
        warnings.warn(
            f'The covariates {tvar} are time varying, but FREM will only use the '
            f'baseline values.'
        )

    # Covariates with only one baseline value
    unique = (cov_bls.iloc[0] == cov_bls.iloc[1:]).all()
    covset = OrderedSet(covariates)
    unique_covariates = OrderedSet(unique[unique].index)
    covset -= unique_covariates
    if unique_covariates:
        warnings.warn(
            f'The covariates {list(unique_covariates)} have the same baseline value for '
            f'all individuals and has been removed from the analysis.'
        )

    # Remove covariates that are equal
    new_covset = OrderedSet(covset)
    for col1, col2 in itertools.combinations(covset, 2):
        if cov_bls[col1].equals(cov_bls[col2]):
            new_covset.discard(col2)
            warnings.warn(
                f'The baselines for covariates {col1} and {col2} are equal for all '
                f'individuals. {col2} has been removed from the analysis.'
            )

    return list(new_covset)


def update_model3b_for_psn(rundir, ncovs):
    """Function to update model3b from psn

    NOTE: This function lets pharmpy tie in to the PsN workflow
          and is a temporary solution
    """
    model_path = Path(rundir) / 'm1'
    model1b = Model.parse_model(model_path / 'model_1b.mod')
    model3 = Model.parse_model(model_path / 'model_3.mod')
    model3_res = read_modelfit_results(model_path / 'model_3.mod')
    model3b = create_model3b(model1b, model3, model3_res, int(ncovs))
    write_model(model3b, model_path, force=True)



from .config import NONMEMConfiguration, conf
from .results import parse_modelfit_results

r"""
.. list-table:: Options for the nonmem plugin
   :widths: 25 25 50 150
   :header-rows: 1

   * - Option name
     - Default value
     - Type
     - Description
   * - ``parameter_names``
     - ``['basic']``
     - list
     - Naming scheme of NONMEM parameters. Possible settings are "abbr" ($ABBR), "comment", and
       "basic". The order denotes priority order
   * - ``default_nonmem_path``
     - Path()
     - pathlib.Path
     - Full path to the default NONMEM installation director
   * - ``write_etas_in_abbr``
     - ``False``
     - bool
     - Whether to write etas as $ABBR records
"""

__all__ = (
    'parse_modelfit_results',
    'conf',
    'NONMEMConfiguration',
)



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Optional, Sequence, Type, TypeVar

from pharmpy.model import Model
from pharmpy.modeling import update_inits
from pharmpy.tools import rank_models, summarize_errors
from pharmpy.workflows import ModelEntry, ModelfitResults, Results, ToolDatabase

from .funcs import summarize_individuals, summarize_individuals_count_table

if TYPE_CHECKING:
    import numpy as np
    import pandas as pd
else:
    from pharmpy.deps import numpy as np
    from pharmpy.deps import pandas as pd

DataFrame = Any  # NOTE: Should be pd.DataFrame but we want lazy loading

RANK_TYPES = frozenset(('ofv', 'lrt', 'aic', 'bic', 'mbic'))


def update_initial_estimates(model: Model, modelfit_results: Optional[ModelfitResults]):
    if modelfit_results is None:
        return model
    if not modelfit_results.minimization_successful:
        if modelfit_results.termination_cause != 'rounding_errors':
            return model

    try:
        model = update_inits(
            model, modelfit_results.parameter_estimates, move_est_close_to_bounds=True
        )
    except (ValueError, np.linalg.LinAlgError):
        pass
    return model


@dataclass(frozen=True)
class ToolResults(Results):
    summary_tool: Optional[Any] = None
    summary_models: Optional[pd.DataFrame] = None
    summary_individuals: Optional[pd.DataFrame] = None
    summary_individuals_count: Optional[pd.DataFrame] = None
    summary_errors: Optional[pd.DataFrame] = None
    final_model: Optional[Model] = None
    models: Sequence[Model] = ()
    tool_database: Optional[ToolDatabase] = None


T = TypeVar('T', bound=ToolResults)


def create_results(
    res_class: Type[T],
    input_model_entry: ModelEntry,
    base_model_entry: ModelEntry,
    cand_model_entries: Sequence[ModelEntry],
    rank_type: str,
    cutoff: Optional[float],
    bic_type: str = 'mixed',
    strictness: Optional[str] = "minimization_successful or (rounding_errors and sigdigs >= 0.1)",
    **rest,
) -> T:
    summary_tool = summarize_tool(
        cand_model_entries, base_model_entry, rank_type, cutoff, bic_type, strictness
    )
    if rank_type == 'lrt':
        delta_name = 'dofv'
    elif rank_type == 'mbic':
        delta_name = 'dbic'
    else:
        delta_name = f'd{rank_type}'

    base_model = base_model_entry.model
    # FIXME: Temporary until parent_model attribute has been removed, e.g. summarize_individuals fails otherwise
    base_model = base_model.replace(parent_model=base_model.name)
    base_res = base_model_entry.modelfit_results

    # FIXME: Temporary until parent_model attribute has been removed, e.g. summarize_individuals fails otherwise
    cand_models = [
        model_entry.model.replace(parent_model=model_entry.parent.name)
        for model_entry in cand_model_entries
    ]
    cand_res = [model_entry.modelfit_results for model_entry in cand_model_entries]

    summary_individuals, summary_individuals_count = summarize_tool_individuals(
        [base_model] + cand_models,
        [base_res] + cand_res,
        summary_tool['description'],
        summary_tool[delta_name],
    )
    summary_errors = summarize_errors([base_res] + cand_res)

    if summary_tool['rank'].isnull().all():
        best_model = None
    else:
        best_model_name = summary_tool['rank'].idxmin()
        best_model = next(
            filter(lambda model: model.name == best_model_name, cand_models), base_model
        )

    if base_model.name != input_model_entry.model.name:
        models = [base_model] + cand_models
    else:
        # Check if any resulting models exist
        if cand_models:
            models = [base_model] + cand_models
        else:
            models = None

    # FIXME: Remove best_model, input_model, models when there is function to read db
    res = res_class(
        summary_tool=summary_tool,
        summary_individuals=summary_individuals,
        summary_individuals_count=summary_individuals_count,
        summary_errors=summary_errors,
        final_model=best_model,
        models=models,
        **rest,
    )

    return res


def summarize_tool(
    model_entries: Sequence[ModelEntry],
    start_model_entry: ModelEntry,
    rank_type: str,
    cutoff: Optional[float],
    bic_type: str = 'mixed',
    strictness: Optional[str] = None,
) -> DataFrame:
    start_model_res = start_model_entry.modelfit_results
    models_res = [model_entry.modelfit_results for model_entry in model_entries]

    if rank_type == 'mbic':
        rank_type = 'bic'
        if len(model_entries) > 0:
            multiple_testing = True
            n_expected_models = len(model_entries)
        else:  # This can happen if the search space of e.g. modelsearch only includes the base model
            multiple_testing = False
            n_expected_models = None
    else:
        multiple_testing = False
        n_expected_models = None

    start_model = start_model_entry.model
    models = [model_entry.model for model_entry in model_entries]

    df_rank = rank_models(
        start_model,
        start_model_res,
        models,
        models_res,
        strictness=strictness,
        rank_type=rank_type,
        cutoff=cutoff,
        bic_type=bic_type,
        multiple_testing=multiple_testing,
        mult_test_p=n_expected_models,
    )
    if rank_type != "lrt" and df_rank.dropna(subset=rank_type).shape[0] == 0:
        raise ValueError("All models fail the strictness criteria!")

    rows = {}

    for model_entry in [start_model_entry] + model_entries:
        model = model_entry.model
        parent_model = model_entry.parent if model_entry.parent is not None else model
        description = model.description
        n_params = len(model.parameters.nonfixed)
        if model.name == start_model.name:
            d_params = 0
        else:
            d_params = n_params - len(parent_model.parameters.nonfixed)
        rows[model.name] = (description, n_params, d_params, parent_model.name)

    colnames = ['description', 'n_params', 'd_params', 'parent_model']
    index = pd.Index(rows.keys(), name='model')
    df_descr = pd.DataFrame(rows.values(), index=index, columns=colnames)

    df = pd.concat([df_descr, df_rank], axis=1)
    df['parent_model'] = df.pop('parent_model')

    df_sorted = df.reindex(df_rank.index)

    assert df_sorted is not None
    return df_sorted


def summarize_tool_individuals(
    models: Sequence[Model],
    models_res: Sequence[ModelfitResults],
    description_col: pd.Series,
    rank_type_col: pd.Series,
):
    summary_individuals = summarize_individuals(models, models_res)
    summary_individuals = summary_individuals.join(description_col, how='inner')
    col_to_move = summary_individuals.pop('description')
    summary_individuals.insert(0, 'description', col_to_move)

    suminds_count = summarize_individuals_count_table(df=summary_individuals)
    if suminds_count is None:
        summary_individuals_count = None
    else:
        suminds_count.insert(0, description_col.name, description_col)
        suminds_count.insert(1, rank_type_col.name, rank_type_col)
        suminds_count['parent_model'] = suminds_count.pop('parent_model')
        summary_individuals_count = suminds_count.sort_values(
            by=[rank_type_col.name], ascending=False
        )
    return summary_individuals, summary_individuals_count



from pharmpy.internals.parse import AttrToken, AttrTree

from .record import Record

_ws = {' ', '\x00', '\t'}


class ProblemRecord(Record):
    @property
    def title(self):
        return str(self.root.subtree('raw_title'))

    def set_title(self, new_title):
        if new_title and new_title[0] in _ws:
            raise ValueError(
                f'Invalid title "{new_title}". Title cannot start with any of {tuple(map(repr, sorted(_ws)))}.'
            )

        title_tree = AttrTree.create('raw_title', dict(ANYTHING=new_title))

        _, _, after = self.root.partition('raw_title')

        root = AttrTree(
            self.root.rule,
            (
                AttrToken('WS', ' '),
                title_tree,
            )
            + after,
        )
        return ProblemRecord(self.name, self.raw_name, root)



from pharmpy.internals.fs.cwd import chdir
from pharmpy.modeling import convert_model, create_basic_pk_model  # , filter_dataset

# from pharmpy.tools import fit, run_structsearch


def test_pkpd(tmp_path, load_model_for_test, testdata):
    with chdir(tmp_path):
        model = create_basic_pk_model('iv', dataset_path=testdata / "nonmem" / "pheno_pd.csv")
        model = convert_model(model, 'nonmem')
    #    pk_model = filter_dataset(model, "DVID != 2")  # NOTE: This step needs to be removed later
    #    pk_res = fit(pk_model)
    #    res = run_structsearch(type='pkpd', search_space="DIRECTEFFECT(*)",results=pk_res, model=model)

    #    no_of_models = 9
    #    assert len(res.summary_models) == no_of_models + 1
    #    assert len(res.summary_tool) == no_of_models + 1
    #    assert len(res.models) == no_of_models

    #    rundir = tmp_path / 'structsearch_dir1'
    #    assert rundir.is_dir()
    #    assert (rundir / 'results.json').exists()
    #    assert (rundir / 'results.csv').exists()
    #    assert (rundir / 'metadata.json').exists()



"""
:meta private:
"""
from typing import Literal, Optional

from pharmpy.deps import sympy
from pharmpy.internals.expr.funcs import PHI
from pharmpy.internals.fn.type import check_list
from pharmpy.model import Assignment, EstimationSteps, JointNormalDistribution, Model

from .data import remove_loq_data, set_lloq_data
from .expressions import _simplify_expression_from_parameters, create_symbol

SUPPORTED_METHODS = frozenset(['m1', 'm3', 'm4', 'm5', 'm6', 'm7'])


def transform_blq(
    model: Model,
    method: Literal['m1', 'm3', 'm4', 'm5', 'm6', 'm7'] = 'm4',
    lloq: Optional[float] = None,
):
    """Transform for BLQ data

    Transform a given model, methods available are m1, m3, m4, m5, m6 and m7 [1]_.
    The blq information can come from the dataset, the lloq option or a combination. Both LLOQ and BLQ
    columns are supported. The table below explains which columns are used for the various cases:

    +-------------+-------------+------------+-------------------+---------------+-------------------+
    | lloq option | LLOQ column | BLQ column | Used as indicator | Used as level | Note              |
    +=============+=============+============+===================+===============+===================+
    | Available   | NA          | NA         | DV < lloq         | lloq          |                   |
    +-------------+-------------+------------+-------------------+---------------+-------------------+
    | NA          | Available   | NA         | DV < LLOQ         | LLOQ          |                   |
    +-------------+-------------+------------+-------------------+---------------+-------------------+
    | NA          | NA          | Available  | BLQ               | nothing       | Only for M1 and M7|
    +-------------+-------------+------------+-------------------+---------------+-------------------+
    | NA          | NA          | NA         | NA                | NA            | No BLQ handling   |
    +-------------+-------------+------------+-------------------+---------------+-------------------+
    | NA          | Available   | Available  | BLQ               | LLOQ          | DV column not used|
    +-------------+-------------+------------+-------------------+---------------+-------------------+
    | Available   | NA          | Available  | BLQ               | lloq          |                   |
    +-------------+-------------+------------+-------------------+---------------+-------------------+
    | Available   | Available   | NA         | DV < lloq         | lloq          | Column overridden |
    +-------------+-------------+------------+-------------------+---------------+-------------------+
    | Available   | Available   | Available  | DV < lloq         | lloq          | Columns overridden|
    +-------------+-------------+------------+-------------------+---------------+-------------------+

    BLQ observations are defined as shown in the table above.
    If both a BLQ and an LLOQ column exist in the dataset and no lloq is specified then all dv values in
    rows with BLQ = 1 are counted as BLQ observations. If instead an lloq value is specified then all rows with
    dv values below the lloq value are counted as BLQ observations.
    If no lloq is specified and no BLQ column exists in the dataset then all rows with dv values below the value
    specified in the DV column are counted as BLQ observations.


    M1 method:
        All BLQ observations are discarded.
        This may affect the size of the dataset.
    M3 method:
        Including the probability that the BLQ observations are below the LLOQ
        as part of the maximum likelihood estimation.
        For more details see :ref:`[1]<ref_article>`.
        This method modifies the Y statement of the model (see examples below).
    M4 method:
        Including the probability that the BLQ observations are below the LLOQ and positive
        as part of the maximum likelihood estimation.
        For more details see :ref:`[1]<ref_article>`.
        This method modifies the Y statement of the model (see examples below).
    M5 method:
        All BLQ observations are replaced by level/2, where level = lloq if lloq is specified.
        Else level = value specified in LLOQ column (see table above).
        This method may change entries in the dataset.
    M6 method:
        Every BLQ observation in a consecutive series of BLQ observations is discarded except for the first one.
        The remaining BLQ observations are replaced by level/2, where level = lloq if lloq is specified.
        Else level = value specified in LLOQ column (see table above).
        This method may change entries in the dataset as well as the size of the dataset.
    M7 method:
        All BLQ observations are replaced by 0.
        This method may change entries in the dataset.



    Current limitations of the m3 and m4 method:

    * Does not support covariance between epsilons
    * Supports additive, proportional, combined, and power error model

    .. _ref_article:

    .. [1] Beal SL. Ways to fit a PK model with some data below the quantification
       limit. J Pharmacokinet Pharmacodyn. 2001 Oct;28(5):481-504. doi: 10.1023/a:1012299115260.
       Erratum in: J Pharmacokinet Pharmacodyn 2002 Jun;29(3):309. PMID: 11768292.

    Parameters
    ----------
    model : Model
        Pharmpy model
    method : str
        Which BLQ method to use
    lloq : float, optional
        LLOQ limit to use, if None Pharmpy will use the BLQ/LLOQ column in the dataset

    Return
    ------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> model = transform_blq(model, method='m4', lloq=0.1)
    >>> model.statements.find_assignment("Y")
        ⎧ EPS₁⋅W + F   for DV ≥ LLOQ
        ⎪
        ⎨CUMD - CUMDZ
        ⎪────────────   otherwise
    Y = ⎩ 1 - CUMDZ

    See also
    --------
    remove_loq_data
    set_lloq_data

    """
    method = method.lower()
    check_list("method", method, SUPPORTED_METHODS)

    blq_col, lloq_col = _get_blq_and_lloq_columns(model)
    indicator, indicator_type, level, level_type = _which_indicator_and_level(
        lloq, lloq_col, blq_col, method
    )

    if method == 'm1':
        model = _m1_method(model, indicator, indicator_type)
    elif method in ('m3', 'm4'):
        _verify_model(model, method)
        model = _m3_m4_method(model, indicator, indicator_type, level, level_type, method)
    elif method == 'm5':
        model = _m5_method(model, indicator, indicator_type, level, level_type)
    elif method == 'm6':
        model = _m6_method(model, indicator, indicator_type, level, level_type)
    elif method == 'm7':
        model = _m7_method(model, indicator, indicator_type)

    return model


def _m1_method(model, indicator, tp):
    if tp in ('lloq', 'LLOQ'):
        return remove_loq_data(model, lloq=indicator)
    else:  # tp == 'blq'
        return remove_loq_data(model, blq=indicator)


def _m5_method(model, indicator, tp, level, level_type):
    if level_type == 'LLOQ':
        half = f'{level}/2'
    else:
        half = level / 2
    if tp in ('lloq', 'LLOQ'):
        return set_lloq_data(model, half, lloq=indicator)
    else:  # tp == 'blq'
        return set_lloq_data(model, half, blq=indicator)


def _m6_method(model, indicator, tp, level, level_type):
    if level_type == 'LLOQ':
        half = f'{level}/2'
    else:
        half = level / 2
    if tp in ('lloq', 'LLOQ'):
        model = remove_loq_data(model, lloq=indicator, keep=1)
        return set_lloq_data(model, half, lloq=indicator)
    else:  # tp == 'blq'
        model = remove_loq_data(model, blq=indicator, keep=1)
        return set_lloq_data(model, half, blq=indicator)


def _m7_method(model, indicator, tp):
    if tp in ('lloq', 'LLOQ'):
        return set_lloq_data(model, 0, lloq=indicator)
    else:  # tp == 'blq':
        return set_lloq_data(model, 0, blq=indicator)


def _m3_m4_method(model, indicator, indicator_type, level, level_type, method):
    sset = model.statements

    est_steps = model.estimation_steps
    est_steps_new = EstimationSteps([est_step.replace(laplace=True) for est_step in est_steps])
    model = model.replace(estimation_steps=est_steps_new)

    # FIXME: Handle other DVs?
    y_symb = list(model.dependent_variables.keys())[0]
    y = sset.find_assignment(y_symb)
    ipred = y.expression.subs({rv: 0 for rv in model.random_variables.epsilons.names})

    lloq_symbol = create_symbol(model, 'LLOQ')
    if indicator_type == 'lloq':
        indicator_symb = lloq_symbol
    else:
        indicator_symb = sympy.Symbol(indicator)
    if level_type == 'lloq':
        level_symb = lloq_symbol
    else:
        level_symb = sympy.Symbol(level)

    sd = _get_sd(model, y)
    symb_dv = sympy.Symbol(model.datainfo.dv_column.name)
    symb_fflag = create_symbol(model, 'F_FLAG')
    symb_cumd = create_symbol(model, 'CUMD')

    if indicator_type in ('lloq', 'LLOQ'):
        is_above_lloq = sympy.GreaterThan(symb_dv, indicator_symb)
    else:
        is_above_lloq = sympy.Equality(indicator_symb, 0)

    assignments = [sd]
    if indicator_type == 'lloq' or level_type == 'lloq':
        if indicator_type == 'lloq':
            symbol = indicator_symb
            value = indicator
        else:
            symbol = level_symb
            value = level
        lloq = Assignment(symbol, sympy.Float(value))
        assignments.append(lloq)

    assignments += Assignment(symb_fflag, sympy.Piecewise((0, is_above_lloq), (1, True)))

    cumd = Assignment(symb_cumd, PHI((level_symb - ipred) / sd.symbol))
    if method == 'm3':
        assignments += Assignment(
            y.symbol, sympy.Piecewise((y.expression, is_above_lloq), (cumd.expression, True))
        )
    else:
        assignments += cumd
        symb_cumdz = create_symbol(model, 'CUMDZ')
        assignments += Assignment(symb_cumdz, PHI(-ipred / sd.symbol))

        y_below_lloq = (symb_cumd - symb_cumdz) / (1 - symb_cumdz)
        assignments += Assignment(
            y.symbol, sympy.Piecewise((y.expression, is_above_lloq), (y_below_lloq, True))
        )

    y_idx = sset.find_assignment_index(y.symbol)
    sset_new = sset[:y_idx] + assignments + sset[y_idx + 1 :]
    model = model.replace(statements=sset_new)

    return model.update_source()


def has_blq_transformation(model: Model):
    # FIXME: Make more general
    y_symb = list(model.dependent_variables.keys())[0]
    y = model.statements.error.find_assignment(y_symb)
    if not y:
        raise ValueError(f'Could not find assignment for \'{y_symb}\'')
    y_expr = y.expression
    if not isinstance(y_expr, sympy.Piecewise):
        return False
    for statement, cond in y_expr.args:
        blq_symb, _ = get_blq_symb_and_type(model)
        if blq_symb in cond.free_symbols:
            break
    else:
        return False

    expected_m3 = ['SD', 'F_FLAG']
    expected_m4 = ['SD', 'F_FLAG', 'CUMD', 'CUMDZ']
    return _has_all_expected_symbs(model.statements.error, expected_m3) or _has_all_expected_symbs(
        model.statements.error, expected_m4
    )


def _get_blq_and_lloq_columns(model: Model):
    try:
        blq_datainfo = model.datainfo.typeix['blq']
    except IndexError:
        blq = None
    else:
        blq = blq_datainfo[0].name
    try:
        lloq_datainfo = model.datainfo.typeix['lloq']
    except IndexError:
        lloq = None
    else:
        lloq = lloq_datainfo[0].name
    return blq, lloq


def _which_indicator_and_level(lloq, lloq_col, blq_col, method):
    # Returns indicator, indicator type, level and level_type
    # indicator type can be 'lloq' for a value, 'LLOQ' for a column or 'blq'
    # level_type can be 'lloq' or 'LLOQ'
    if lloq is not None:
        if lloq_col is None and blq_col is not None:
            return blq_col, 'blq', lloq, 'lloq'
        else:
            return lloq, 'lloq', lloq, 'lloq'
    elif blq_col is not None:
        if lloq is None and lloq_col is None and method not in ('m1', 'm7'):
            raise ValueError(
                "Can only find a BLQ column. Only supported for the M1 and M7 methods."
            )
        return blq_col, 'blq', lloq_col, 'LLOQ'
    elif lloq_col is not None:
        return lloq_col, 'LLOQ', lloq_col, 'LLOQ'
    else:
        raise ValueError("No BLQ or LLOQ information available")


def _get_blq_name_and_type(model: Model):
    try:
        blq_datainfo = model.datainfo.typeix['lloq']
        return blq_datainfo[0].name, 'lloq'
    except IndexError:
        blq_datainfo = model.datainfo.typeix['blq']
        return blq_datainfo[0].name, 'blq'


def get_blq_symb_and_type(model: Model):
    try:
        name, tp = _get_blq_name_and_type(model)
        return sympy.Symbol(name), tp
    except IndexError:
        return sympy.Symbol('LLOQ'), 'lloq'


def _has_all_expected_symbs(sset, expected_symbs):
    symb_names = [s.symbol.name for s in sset]
    return all(symb in symb_names for symb in expected_symbs)


def _verify_model(model, method):
    rvs = model.random_variables.epsilons
    if any(isinstance(rv, JointNormalDistribution) for rv in rvs):
        raise ValueError(
            f'Invalid input model: covariance between epsilons not supported in `method` {method}'
        )


def _get_sd(model, y):
    y_expr = model.statements.find_assignment(y.symbol).expression
    sd_expr = get_sd_expr(y_expr, model.random_variables, model.parameters)
    symb_sd = create_symbol(model, 'SD')
    return Assignment(symb_sd, sd_expr)


def get_sd_expr(y_expr, rvs, params):
    rv_terms = [arg for arg in y_expr.args if arg.free_symbols.intersection(rvs.free_symbols)]
    sd_expr = []
    for i, term in enumerate(rv_terms, 1):
        rvs_in_term = rvs.free_symbols.intersection(term.free_symbols)
        if len(rvs_in_term) > 1:
            raise ValueError(
                'Invalid input model: error model not supported, terms in error model cannot contain '
                'more than one random variable'
            )
        expr = rvs.replace_with_sympy_rvs(term)
        sd_expr.append(sympy.stats.std(expr))

    return _simplify_expression_from_parameters(
        sympy.sqrt(sympy.Add(*[expr**2 for expr in sd_expr])), params
    )



from pharmpy.model import Assignment
from pharmpy.modeling import replace_non_random_rvs


def test_replace_non_random_rvs(create_model_for_test):
    code = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV FA1 FA2
$SUBROUTINE ADVAN1 TRANS2
$PK
CL=THETA(1)*EXP(ETA(1))
VC=THETA(2)*EXP(ETA(2))+OMEGA(2,2)
V=VC
$ERROR
Y=F+F*EPS(1)
$THETA (0,0.00469307) ; POP_CL
$THETA (0,0.22) ; POP_VC
$OMEGA 0.0309626  ; IVCL
$OMEGA 0 FIX  ; IVV
$SIGMA 0.013241 ; SIGMA
$ESTIMATION METHOD=1 INTERACTION
"""
    model = create_model_for_test(code, dataset='pheno')
    new = replace_non_random_rvs(model)
    assert new.parameters == model.parameters[['POP_CL', 'POP_VC', 'IVCL', 'SIGMA']]
    assert new.random_variables == model.random_variables[['ETA_1', 'EPS_1']]
    assert new.statements[0] == Assignment.create("CL", "POP_CL*exp(ETA_1)")
    assert new.statements[1] == Assignment.create("VC", "POP_VC")



# The modeling features language
# A high level language to describe model features and ranges of model features

# whitespace between tokens are ignored
# ; and \n separate features
# feature names are case insensitive
# FEATURE_NAME(options)     options is a comma separated list
#
# absorption(x)      - x can be FO, ZO, SEQ-ZO-FO, [FO, ZO] for multiple or * for all
# elimination(x)     - x can be FO, ZO, MM and MIX-FO-MM
# peripherals(n)     - n can be 0, 1, ... or a..b for a range (inclusive in both ends)
# transits(n[, d])   - n as above, d (optional) is DEPOT, NODEPOT, or *
# covariate(p, c, e) - p is a parameter name, a list of parameter names, or a reference
#                    - c is a covariate name, a list of covariate names, or a reference
#                    - e is an effect, a list of effects, or * for all continuous effects
#
# LET(variable_name, values) defines a new variable to be a value or list of values


grammar = r"""
%ignore " "
_SEPARATOR: /;|\n/
start: _statement (_SEPARATOR _statement)*
_statement: definition | _feature

definition: "LET"i "(" VARIABLE_NAME "," values ")"

_feature: absorption | elimination | peripherals | transits | lagtime
            | covariate | direct_effect | effect_comp | indirect_effect
            | metabolite

absorption: "ABSORPTION"i "(" (_absorption_option) ")"
elimination: "ELIMINATION"i "(" (_elimination_option) ")"
peripherals: "PERIPHERALS"i "(" (_counts) ["," _peripheral_comp] ")"
transits: "TRANSITS"i "(" _counts ["," _depot_option] ")"
lagtime: "LAGTIME"i "(" (_lagtime_option) ")"
covariate: "COVARIATE"i [optional_cov] "(" parameter_option "," covariate_option "," fp_option ["," op_option] ")"

direct_effect: "DIRECTEFFECT"i "(" (_pdtype_option) ")"
effect_comp: "EFFECTCOMP"i "(" (_pdtype_option) ")"
indirect_effect: "INDIRECTEFFECT"i "(" _pdtype_option "," _production_option ")"

metabolite: "METABOLITE"i "(" (_metabolite_option) ")"

_pdtype_option: pdtype_modes | pdtype_wildcard
pdtype_modes: PDTYPE_MODE | "[" [PDTYPE_MODE ("," PDTYPE_MODE)*] "]"
PDTYPE_MODE: "linear"i | "Emax"i | "sigmoid"i
pdtype_wildcard: WILDCARD

_production_option: production_modes | production_wildcard
production_modes: PRODUCTION_MODE
PRODUCTION_MODE: "production"i | "degradation"i
production_wildcard: WILDCARD

_metabolite_option: metabolite_modes | metabolite_wildcard
metabolite_modes: METABOLITE_MODE | "[" [METABOLITE_MODE ("," METABOLITE_MODE)*] "]"
METABOLITE_MODE: "basic"i | "psc"i
metabolite_wildcard: WILDCARD

_absorption_option: absorption_modes | absorption_wildcard
absorption_modes: ABSORPTION_MODE | "[" [ABSORPTION_MODE ("," ABSORPTION_MODE)*] "]"
absorption_wildcard: WILDCARD
ABSORPTION_MODE: "FO"i | "ZO"i | "SEQ-ZO-FO"i | "INST"i

_elimination_option: elimination_modes | elimination_wildcard
elimination_modes: ELIMINATION_MODE | "[" [ELIMINATION_MODE ("," ELIMINATION_MODE)*] "]"
elimination_wildcard: WILDCARD
ELIMINATION_MODE: "FO"i | "ZO"i | "MM"i | "MIX-FO-MM"i

_depot_option: depot_modes | depot_wildcard
depot_modes: DEPOT_MODE | "[" [DEPOT_MODE ("," DEPOT_MODE)*] "]"
DEPOT_MODE: "DEPOT"i |"NODEPOT"i
depot_wildcard: WILDCARD

_peripheral_comp: peripheral_modes | peripheral_wildcard
peripheral_modes: PERIPHERAL_MODE | "[" [PERIPHERAL_MODE ("," PERIPHERAL_MODE)*] "]"
PERIPHERAL_MODE: "DRUG"i | "MET"i
peripheral_wildcard: WILDCARD

_lagtime_option: lagtime_modes | lagtime_wildcard
lagtime_modes: LAGTIME_MODE | "[" [LAGTIME_MODE ("," LAGTIME_MODE)*] "]"
lagtime_wildcard: WILDCARD
LAGTIME_MODE: "ON"i | "OFF"i

parameter_option: values | ref | parameter_wildcard
covariate_option: values | ref | covariate_wildcard
fp_option: values | fp_wildcard
!op_option: "+" | "*"
optional_cov: OPTIONAL

ref: "@" VARIABLE_NAME
parameter_wildcard: WILDCARD
covariate_wildcard: WILDCARD
fp_wildcard: WILDCARD

WILDCARD: "*"
OPTIONAL: "?"

VARIABLE_NAME: /[a-zA-Z_]+/

values: value | _value_array
_value_array: "[" [value ("," value)*] "]"
value: /[a-zA-Z0-9-]+/

_counts: count | count_array
count_array: "[" [number ("," number)*] "]"
count: number | range
range: number ".." number
number: /\d+/
"""



from threading import Lock

from .mfl.parse import get_model_features

__all__ = (
    'create_report',  # pyright: ignore [reportUnsupportedDunderAll]
    'create_results',  # pyright: ignore [reportUnsupportedDunderAll]
    'fit',  # pyright: ignore [reportUnsupportedDunderAll]
    'is_strictness_fulfilled',  # pyright: ignore [reportUnsupportedDunderAll]
    'load_example_modelfit_results',  # pyright: ignore [reportUnsupportedDunderAll]
    'predict_influential_individuals',  # pyright: ignore [reportUnsupportedDunderAll]
    'predict_influential_outliers',  # pyright: ignore [reportUnsupportedDunderAll]
    'predict_outliers',  # pyright: ignore [reportUnsupportedDunderAll]
    'print_fit_summary',  # pyright: ignore [reportUnsupportedDunderAll]
    'rank_models',  # pyright: ignore [reportUnsupportedDunderAll]
    'read_results',  # pyright: ignore [reportUnsupportedDunderAll]
    'read_modelfit_results',  # pyright: ignore [reportUnsupportedDunderAll]
    'resume_tool',  # pyright: ignore [reportUnsupportedDunderAll]
    'retrieve_final_model',  # pyright: ignore [reportUnsupportedDunderAll]
    'retrieve_models',  # pyright: ignore [reportUnsupportedDunderAll]
    'run_allometry',  # pyright: ignore [reportUnsupportedDunderAll]
    'run_amd',  # pyright: ignore [reportUnsupportedDunderAll]
    'run_bootstrap',  # pyright: ignore [reportUnsupportedDunderAll]
    'run_covsearch',  # pyright: ignore [reportUnsupportedDunderAll]
    'run_estmethod',  # pyright: ignore [reportUnsupportedDunderAll]
    'run_iivsearch',  # pyright: ignore [reportUnsupportedDunderAll]
    'run_iovsearch',  # pyright: ignore [reportUnsupportedDunderAll]
    'run_modelfit',  # pyright: ignore [reportUnsupportedDunderAll]
    'run_modelsearch',  # pyright: ignore [reportUnsupportedDunderAll]
    'run_retries',  # pyright: ignore [reportUnsupportedDunderAll]
    'run_ruvsearch',  # pyright: ignore [reportUnsupportedDunderAll]
    'run_structsearch',  # pyright: ignore [reportUnsupportedDunderAll]
    'run_tool',  # pyright: ignore [reportUnsupportedDunderAll]
    'summarize_errors',  # pyright: ignore [reportUnsupportedDunderAll]
    'summarize_individuals',  # pyright: ignore [reportUnsupportedDunderAll]
    'summarize_individuals_count_table',  # pyright: ignore [reportUnsupportedDunderAll]
    'summarize_modelfit_results',  # pyright: ignore [reportUnsupportedDunderAll]
    'write_results',  # pyright: ignore [reportUnsupportedDunderAll]
    'get_model_features',
)


_allowed = set(__all__)

_not_wrapped = {
    '.amd.run': ('run_amd',),
    '.reporting': ('create_report',),
    '.run': (
        'create_results',
        'fit',
        'is_strictness_fulfilled',
        'load_example_modelfit_results',
        'print_fit_summary',
        'rank_models',
        'read_modelfit_results',
        'read_results',
        'resume_tool',
        'retrieve_final_model',
        'retrieve_models',
        'run_tool',
        'summarize_errors',
        'summarize_modelfit_results',
        'write_results',
    ),
    '.funcs': (
        'predict_outliers',
        'predict_influential_individuals',
        'predict_influential_outliers',
        'summarize_individuals',
        'summarize_individuals_count_table',
    ),
}

_not_wrapped_module_name_index = {
    key: module for module, keys in _not_wrapped.items() for key in keys
}

_tool_cache = {}
_tool_lock = Lock()


def __getattr__(key):
    if key not in _allowed:
        raise AttributeError(key)

    import importlib

    if key in _not_wrapped_module_name_index:
        module = importlib.import_module(_not_wrapped_module_name_index[key], __name__)
        return getattr(module, key)

    assert key.startswith('run_')

    with _tool_lock:
        if key not in _tool_cache:
            module = importlib.import_module('.wrap', __name__)
            wrap = getattr(module, 'wrap')
            tool_name = key[4:]  # NOTE: This removes the run_ prefix
            _tool_cache[key] = wrap(tool_name)

    return _tool_cache[key]


def __dir__():
    return __all__



import inspect

from .workflow import WorkflowBuilder


def insert_context(wb: WorkflowBuilder, context):
    """Insert tool context (database) for all tasks in a workflow needing it

    having context as first argument of function
    """
    for task in wb.tasks:
        parameters = tuple(inspect.signature(task.function).parameters)
        if parameters and parameters[0] == 'context':
            new_task = task.replace(task_input=(context, *task.task_input))
            wb.replace_task(task, new_task)
